#cython: language_level=3

import numpy as np

cimport cython

cdef int ENCODING_LEN = 4


@cython.boundscheck(False)
@cython.wraparound(False)
def compute_encoded_kmer_batch(
    int before_context_bases,
    int after_context_bases,
    const signed char[:, ::1] seqs,
    const short[:, ::1] seq_mappings,
    const short[::1] seq_lens,
):
    cdef int nchunks = seq_lens.shape[0]

    # initialize output array
    cdef int sig_len = seq_mappings[0, seq_lens[0]]
    cdef int kmer_len = before_context_bases + after_context_bases + 1
    cdef int enc_kmer_len = ENCODING_LEN * kmer_len
    out_arr = np.zeros((seq_lens.shape[0], enc_kmer_len, sig_len), np.float32)
    cdef float[:, :, ::1] out_mv = out_arr

    # loop over chunks, kmer_pos and mappings to fill output array
    cdef int chunk_idx, seq_len, kmer_pos, enc_offset
    cdef int seq_pos, base, base_st, base_en, sig_pos
    for chunk_idx in range(nchunks):
        seq_len = seq_lens[chunk_idx]
        for kmer_pos in range(kmer_len):
            enc_offset = ENCODING_LEN * kmer_pos
            for seq_pos in range(seq_len):
                base = seqs[chunk_idx, seq_pos + kmer_pos]
                if base == -1:
                    continue
                base_st = seq_mappings[chunk_idx, seq_pos]
                base_en = seq_mappings[chunk_idx, seq_pos + 1]
                for sig_pos in range(base_st, base_en):
                    out_mv[chunk_idx, enc_offset + base, sig_pos] = 1.0
    return out_arr



import os
import re
import array
import queue
import platform
import traceback
from time import sleep
import multiprocessing as mp
from threading import Thread
from os.path import realpath, expanduser

import torch
import numpy as np

from remora import log, RemoraError

LOGGER = log.get_logger()

CAN_ALPHABET = "ACGT"
CONV_ALPHABET = "ACGTN"
SINGLE_LETTER_CODE = {
    "A": "A",
    "C": "C",
    "G": "G",
    "T": "T",
    "B": "CGT",
    "D": "AGT",
    "H": "ACT",
    "K": "GT",
    "M": "AC",
    "N": "ACGT",
    "R": "AG",
    "S": "CG",
    "V": "ACG",
    "W": "AT",
    "Y": "CT",
}
SEQ_MIN = np.array(["A"], dtype="S1").view(np.uint8)[0]
SEQ_TO_INT_ARR = np.full(26, -1, dtype=int)
SEQ_TO_INT_ARR[0] = 0
SEQ_TO_INT_ARR[2] = 1
SEQ_TO_INT_ARR[6] = 2
SEQ_TO_INT_ARR[19] = 3
COMP_BASES = dict(zip(map(ord, "ACGT"), map(ord, "TGCA")))
NP_COMP_BASES = np.array([3, 2, 1, 0], dtype=np.uintp)
U_TO_T_BASES = {ord("U"): ord("T")}
T_TO_U_BASES = {ord("T"): ord("U")}

DEFAULT_QUEUE_SIZE = 10_000


def parse_device(device):
    if device is None:
        return None
    # convert int devices for legacy settings
    try:
        device = int(device)
    except (ValueError, TypeError):
        pass
    device = torch.device(device)
    if device.type == "cuda" and not torch.cuda.is_available():
        LOGGER.warning("Device option specified, but CUDA not available.")
    return device


def iter_motif_hits(int_seq, motif):
    yield from np.where(
        np.logical_and.reduce(
            [
                np.isin(
                    int_seq[
                        po : int_seq.size - len(motif.int_pattern) + po + 1
                    ],
                    pi,
                )
                for po, pi in enumerate(motif.int_pattern)
            ]
        )
    )[0]


def find_focus_bases_in_int_sequence(
    int_seq: np.ndarray, motifs: list
) -> np.ndarray:
    return np.fromiter(
        set(
            mot_pos + mot.focus_pos
            for mot in motifs
            for mot_pos in iter_motif_hits(int_seq, mot)
        ),
        int,
    )


def comp(seq):
    return seq.translate(COMP_BASES)


def revcomp(seq):
    return seq.upper().translate(COMP_BASES)[::-1]


def comp_np(np_seq):
    return NP_COMP_BASES[np_seq]


def revcomp_np(np_seq):
    return NP_COMP_BASES[np_seq][::-1]


def u_to_t(seq):
    return seq.translate(U_TO_T_BASES)


def t_to_u(seq):
    return seq.translate(T_TO_U_BASES)


def seq_to_int(seq):
    """Convert string sequence to integer encoded array

    Args:
        seq (str): Nucleotide sequence

    Returns:
        np.array containing integer encoded sequence
    """
    return SEQ_TO_INT_ARR[
        np.array(list(seq), dtype="c").view(np.uint8) - SEQ_MIN
    ]


def int_to_seq(np_seq, alphabet=CONV_ALPHABET):
    """Convert integer encoded array to string sequence

    Args:
        np_seq (np.array): integer encoded sequence

    Returns:
        String nucleotide sequence
    """
    if np_seq.shape[0] == 0:
        return ""
    if np_seq.max() >= len(alphabet):
        raise RemoraError(f"Invalid value in int sequence ({np_seq.max()})")
    return "".join(alphabet[b] for b in np_seq)


def resolve_path(fn_path):
    """Helper function to resolve relative and linked paths that might
    give other packages problems.
    """
    if fn_path is None:
        return None
    return realpath(expanduser(fn_path))


def to_str(value):
    """Try to convert a bytes object to a string. If it is already a string
    catch this error and return the input string. This can be used for read ids
    stored in HDF5 files as they are sometimes returned as bytes and sometimes
    strings.
    """
    try:
        return value.decode()
    except AttributeError:
        return value


def softmax_axis1(x):
    """Compute softmax over axis=1"""
    e_x = np.exp((x.T - np.max(x, axis=1)).T)
    with np.errstate(divide="ignore"):
        return (e_x.T / e_x.sum(axis=1)).T


def get_read_ids(bam_idx, pod5_fh, num_reads):
    """Get overlapping read ids from bam index and pod5 file

    Args:
        bam_idx (ReadIndexedBam): Read indexed BAM
        pod5_fh (pod5.Reader): POD5 file handle
        num_reads (int): Maximum number of reads, or None for no max
    """
    LOGGER.info("Extracting read IDs from POD5")
    pod5_read_ids = set(pod5_fh.read_ids)
    num_pod5_reads = len(pod5_read_ids)
    # pod5 will raise when it cannot find a "selected" read id, so we make
    # sure they're all present before starting
    # todo(arand) this could be performed using the read_table instead, but
    #  it's worth checking that it's actually faster and doesn't explode
    #  memory before switching from a sweep throug the pod5 file
    both_read_ids = list(pod5_read_ids.intersection(bam_idx.read_ids))
    num_both_read_ids = len(both_read_ids)
    LOGGER.info(
        f"Found {bam_idx.num_reads} BAM records, {num_pod5_reads} "
        f"POD5 reads, and {num_both_read_ids} in common"
    )
    if num_reads is None:
        num_reads = num_both_read_ids
    else:
        num_reads = min(num_reads, num_both_read_ids)
    return both_read_ids, num_reads


class Motif:
    def __init__(self, raw_motif, focus_pos=0):
        self.raw_motif = raw_motif
        try:
            self.focus_pos = int(focus_pos)
        except ValueError:
            raise RemoraError(
                f'Motif focus position not an integer: "{focus_pos}"'
            )
        if self.focus_pos >= len(self.raw_motif):
            raise RemoraError(
                "Motif focus position is past the end of the motif"
            )

        ambig_pat_str = "".join(
            "[{}]".format(SINGLE_LETTER_CODE[letter]) for letter in raw_motif
        )
        # add lookahead group to serach for overlapping motif hits
        self.pattern = re.compile("(?=({}))".format(ambig_pat_str))

        self.int_pattern = [
            np.array(
                [
                    b_idx
                    for b_idx, b in enumerate(CAN_ALPHABET)
                    if b in SINGLE_LETTER_CODE[letter]
                ]
            )
            for letter in raw_motif
        ]

    def to_tuple(self):
        return self.raw_motif, self.focus_pos

    @property
    def focus_base(self):
        return self.raw_motif[self.focus_pos]

    @property
    def any_context(self):
        return self.raw_motif == "N"

    @property
    def num_bases_after_focus(self):
        return len(self.raw_motif) - self.focus_pos - 1


def get_can_converter(alphabet, collapse_alphabet):
    """Compute conversion from full alphabet integer encodings to
    canonical alphabet integer encodings.
    """
    can_bases = "".join(
        (
            can_base
            for mod_base, can_base in zip(alphabet, collapse_alphabet)
            if mod_base == can_base
        )
    )
    return np.array(
        [can_bases.find(b) for b in collapse_alphabet], dtype=np.byte
    )


def get_mod_bases(alphabet, collapse_alphabet):
    return [
        mod_base
        for mod_base, can_base in zip(alphabet, collapse_alphabet)
        if mod_base != can_base
    ]


def validate_mod_bases(
    mod_bases, motifs, alphabet, collapse_alphabet, control=False
):
    """Validate that inputs are mutually consistent. Return label conversion
    from alphabet integer encodings to modified base categories.
    """
    if len(set(mod_bases)) < len(mod_bases):
        raise RemoraError("Single letter modified base codes must be unique.")
    can_base = motifs[0].focus_base
    if any(mot.focus_base != can_base for mot in motifs):
        raise RemoraError(
            "All motifs must be alternatives to the same canonical base"
        )
    can_base_idx = alphabet.find(can_base)
    label_conv = np.full(len(alphabet), -1, dtype=np.byte)
    label_conv[can_base_idx] = 0
    if control:
        return label_conv
    for mod_base in mod_bases:
        if mod_base not in alphabet:
            raise RemoraError("Modified base provided not found in alphabet")
        mod_can_equiv = collapse_alphabet[alphabet.find(mod_base)]
        if mod_can_equiv != can_base:
            raise RemoraError(
                f"Canonical base within motif ({can_base}) does not match "
                f"canonical equivalent for modified base ({mod_can_equiv})"
            )
    for mod_i, mod_base in enumerate(mod_bases):
        label_conv[alphabet.find(mod_base)] = mod_i + 1
    return label_conv


def format_mm_ml_tags(seq, poss, probs, mod_bases, can_base, strand: str = "+"):
    """Format MM and ML tags for BAM output. See
    https://samtools.github.io/hts-specs/SAMtags.pdf for format details.

    Args:
        seq (str): read-centric read sequence. For reference-anchored calls
            this should be the reverse complement sequence.
        poss (list): positions relative to seq
        probs (np.array): probabilities for modified bases
        mod_bases (str): modified base single letter codes
        can_base (str): canonical base
        strand (bool): should be '+' for SEQ-oriented strand and '-' if
            complement strand

    Returns:
        MM string tag and ML array tag
    """

    # initialize dict with all called mods to make sure all called mods are
    # shown in resulting tags
    per_mod_probs = dict((mod_base, []) for mod_base in mod_bases)
    for pos, mod_probs in sorted(zip(poss, probs)):
        # mod_probs is set to None if invalid sequence is encountered or too
        # few events are found around a mod
        if mod_probs is None:
            continue
        for mod_prob, mod_base in zip(mod_probs, mod_bases):
            per_mod_probs[mod_base].append((pos, mod_prob))

    mm_tag, ml_tag = "", array.array("B")
    for mod_base, pos_probs in per_mod_probs.items():
        if len(pos_probs) == 0:
            continue
        mod_poss, probs = zip(*sorted(pos_probs))
        # compute modified base positions relative to the running total of the
        # associated canonical base
        can_base_mod_poss = (
            np.cumsum([1 if b == can_base else 0 for b in seq])[
                np.array(mod_poss)
            ]
            - 1
        )
        mod_gaps = ",".join(
            map(str, np.diff(np.insert(can_base_mod_poss, 0, -1)) - 1)
        )
        mm_tag += f"{can_base}{strand}{mod_base}?,{mod_gaps};"
        # extract mod scores and scale to 0-255 range
        scaled_probs = np.floor(np.array(probs) * 256)
        # last interval includes prob=1
        scaled_probs[scaled_probs == 256] = 255
        ml_tag.extend(scaled_probs.astype(np.uint8))

    return mm_tag, ml_tag


###################
# Multiprocessing #
###################


def _put_item(item, out_q):
    """Put item into queue with timeout to handle KeyboardInterrupt"""
    while True:
        try:
            return out_q.put(item, timeout=0.1)
        except queue.Full:
            continue


def _get_item(in_q):
    """Get item from queue with timeout to handle KeyboardInterrupt"""
    while True:
        try:
            return in_q.get(timeout=0.1)
        except queue.Empty:
            continue


def _queue_iter(in_q, num_proc=1):
    comp_proc = 0
    while comp_proc < num_proc:
        item = _get_item(in_q)
        if item is StopIteration:
            comp_proc += 1
        else:
            yield item


def _fill_q(iterator, in_q, num_recievers):
    try:
        for item in iterator:
            _put_item(item, in_q)
    except KeyboardInterrupt:
        pass
    for _ in range(num_recievers):
        _put_item(StopIteration, in_q)


def _mt_func(func, in_q, out_q, prep_func, name, *args, **kwargs):
    LOGGER.debug(f"Starting {name} worker")
    try:
        if prep_func is not None:
            args, kwargs = prep_func(*args, **kwargs)
        for val in _queue_iter(in_q):
            try:
                out_q.put(func(val, *args, **kwargs))
            except Exception as e:
                # avoid killing thread for a python error.
                LOGGER.debug(
                    f"UNEXPECTED_ERROR in {name} worker: '{e}'.\n"
                    f"Full traceback: {traceback.format_exc()}"
                )
            except KeyboardInterrupt:
                LOGGER.debug(f"stopping {name} due to user interrupt")
                break

    except KeyboardInterrupt:
        pass
    except Exception as e:
        LOGGER.debug(
            f"UNEXPECTED_ERROR in {name} worker: '{e}'.\n"
            f"Full traceback: {traceback.format_exc()}"
        )
    LOGGER.debug(f"Completed {name} worker")
    out_q.put(StopIteration)


def _background_filler(func, args, kwargs, out_q, name):
    LOGGER.debug(f"Starting {name} background filler")
    try:
        for item in func(*args, **kwargs):
            _put_item(item, out_q)
    except KeyboardInterrupt:
        pass
    except Exception as e:
        LOGGER.debug(
            f"UNEXPECTED_ERROR in {name} worker: '{e}'.\n"
            f"Full traceback: {traceback.format_exc()}"
        )
    LOGGER.debug(f"Completed {name} background filler")
    out_q.put(StopIteration)


class MultitaskMap:
    """Map a function to an input queue using multitasking (multiprocessing or
    threading)

    MultitaskMap supports multiprocessing or threading backends via the
    use_process argument.

    Elements of the input queue will be passed as the first argument to the
    func followed by args and kwargs provided.

    MultitaskMap also supports a prepare function to perform work on the input
    arguments within the newly spawned task. The prep_func should take the args
    and kwargs provided to MultitaskMap and return a new set of args and kwargs
    to be passed to the worker function along with elements from the in_q. This
    can be useful for objects that need initialization within a task.

    MultitaskMap supports KeyboardInterrupt without flooding the output with
    stack traces from each killed task to exit gracefully and avoid stalling.
    """

    def __init__(
        self,
        func,
        iterator,
        prep_func=None,
        num_workers=1,
        q_maxsize=DEFAULT_QUEUE_SIZE,
        use_process=False,
        args=(),
        kwargs={},
        name="MultitaskMap",
    ):
        self.name = name
        self.num_workers = num_workers
        self.out_q = mp.Queue(q_maxsize)
        in_q = mp.Queue(q_maxsize)

        mt_worker = mp.Process if use_process else Thread
        # TODO save workers to self and provide method to watch workers
        #   for failures to avoid deadlock
        mt_worker(
            target=_fill_q,
            args=(iterator, in_q, self.num_workers),
            name=f"{self.name}_filler",
            daemon=True,
        ).start()
        args = [func, in_q, self.out_q, prep_func, self.name] + list(args)
        for idx in range(self.num_workers):
            mt_worker(
                target=_mt_func,
                args=args,
                kwargs=kwargs,
                name=f"{self.name}_{idx}",
                daemon=True,
            ).start()
        # processes take a second to start up on mac
        if platform.system() == "Darwin":
            wait_time = os.environ.get("MP_WAIT_TIME")
            if wait_time is None:
                wait_time = 1
            else:
                try:
                    wait_time = int(wait_time)
                except ValueError as e:
                    raise ValueError(
                        f"failed to interpret MP_WAIT_TIME {wait_time}"
                    ) from e
            LOGGER.debug(
                f"MacOS requires that we wait, set MP_WAIT_TIME to modulate, "
                f"waiting {wait_time}s before starting {self.name}"
            )
            sleep(wait_time)

    def __iter__(self):
        try:
            yield from _queue_iter(self.out_q, self.num_workers)
        except KeyboardInterrupt:
            LOGGER.debug(f"MultitaskMap {self.name} interrupted")
            pass


class BackgroundIter:
    def __init__(
        self,
        func,
        q_maxsize=DEFAULT_QUEUE_SIZE,
        use_process=False,
        args=(),
        kwargs={},
        name="BackgroundIter",
    ):
        self.name = name
        self.out_q = mp.Queue(q_maxsize)

        mt_worker = mp.Process if use_process else Thread
        mt_worker(
            target=_background_filler,
            args=(func, args, kwargs, self.out_q, self.name),
            name=f"{self.name}_filler",
            daemon=True,
        ).start()
        # processes take a second to start up on mac
        if platform.system() == "Darwin":
            sleep(1)

    def __iter__(self):
        try:
            yield from _queue_iter(self.out_q)
        except KeyboardInterrupt:
            LOGGER.debug(f"BackgroundIter {self.name} interrupted")
            pass


if __name__ == "__main__":
    RuntimeError("This is a module.")



import os
import itertools
from copy import copy
import array as pyarray
from pathlib import Path
from collections import defaultdict

import pod5
import pysam
import numpy as np
from tqdm import tqdm
from torch.jit._script import RecursiveScriptModule

from remora import constants, log, RemoraError
from remora.io import (
    ReadIndexedBam,
    iter_signal,
    extract_alignments,
    DuplexRead,
    DuplexPairsBuilder,
)
from remora.util import (
    MultitaskMap,
    BackgroundIter,
    format_mm_ml_tags,
    softmax_axis1,
    get_read_ids,
    Motif,
    revcomp,
)

LOGGER = log.get_logger()
_PROF_PREP_FN = os.getenv("REMORA_INFER_PREP_DATA_PROFILE_FILE")
_PROF_MODEL_FN = os.getenv("REMORA_INFER_RUN_MODEL_PROFILE_FILE")
_PROF_MAIN_FN = os.getenv("REMORA_INFER_MAIN_PROFILE_FILE")


################
# Core Methods #
################


def call_read_mods(
    read,
    model,
    model_metadata,
    batch_size=constants.DEFAULT_BATCH_SIZE,
    focus_offset=None,
    return_mm_ml_tags=False,
    return_mod_probs=False,
):
    """Call modified bases on a read.

    Args:
        read (RemoraRead): Read to be called
        model: Compiled inference model
            (see remora.model_util.load_torchscript_model)
        model_metadata: Inference model metadata
        batch_size (int): Number of chunks to call per-batch
        focus_offset (int): Specific base to call within read
            Default: Use motif from model
        return_mm_ml_tags (bool): Return MM and ML tags for SAM tags.
        return_mod_probs (bool): Convert returned neural network score to
            probabilities

    Returns:
        If return_mm_ml_tags, MM string tag and ML array tag
        Else if return_mod_probs, 3-tuple containing:
          1. Modified base probabilities (dim: num_calls, num_mods)
          2. Labels for each base (-1 if labels not provided)
          3. List of positions within the `read.seq`.
       Else, return value from call_read_mods_core
    """
    if focus_offset is None:
        motifs = [Motif(*mot) for mot in model_metadata["motifs"]]
        read.set_motif_focus_bases(motifs)
    else:
        read.focus_bases = np.array([focus_offset])
    read.prepare_batches(model_metadata, batch_size)
    if len(read.batches) == 0:
        return np.array([]), np.array([]), np.array([])
    nn_out, labels, pos = read.run_model(model)
    if not return_mod_probs and not return_mm_ml_tags:
        return nn_out, labels, pos
    probs = softmax_axis1(nn_out)[:, 1:].astype(np.float64)
    if return_mm_ml_tags:
        return format_mm_ml_tags(
            seq=read.str_seq,
            poss=pos,
            probs=probs,
            mod_bases=model_metadata["mod_bases"],
            can_base=model_metadata["can_base"],
        )
    return probs, labels, pos


################
# Duplex Infer #
################


class DuplexReadModCaller:
    def __init__(self, model, model_metadata):
        self.model = model
        self.model_metadata = model_metadata

    def call_duplex_read_mod_probs(
        self,
        duplex_read: DuplexRead,
    ):
        template_read = duplex_read.template_read.into_remora_read(False)
        complement_read = duplex_read.complement_read.into_remora_read(False)

        template_probs, _, template_positions = call_read_mods(
            read=template_read,
            model=self.model,
            model_metadata=self.model_metadata,
            return_mod_probs=True,
        )
        template_positions = template_positions + duplex_read.template_ref_start

        complement_probs, _, complement_positions = call_read_mods(
            read=complement_read,
            model=self.model,
            model_metadata=self.model_metadata,
            return_mod_probs=True,
        )
        complement_positions = (
            complement_positions + duplex_read.complement_ref_start
        )

        read_sequence = (
            duplex_read.duplex_basecalled_sequence
            if not duplex_read.is_reverse_mapped
            else revcomp(duplex_read.duplex_basecalled_sequence)
        )

        if duplex_read.is_reverse_mapped:
            (template_positions, template_probs), (
                complement_positions,
                complement_probs,
            ) = (complement_positions, complement_probs), (
                template_positions,
                template_probs,
            )

        complement_positions_duplex_orientation = (
            len(read_sequence) - complement_positions - 1
        )

        return {
            "template_probs": template_probs,
            "template_positions": template_positions,
            "complement_probs": complement_probs,
            "complement_positions": complement_positions_duplex_orientation,
            "read_sequence": read_sequence,
        }

    def call_duplex_read_mods(
        self,
        duplex_read: DuplexRead,
    ) -> (str, pyarray.array):
        duplex_read_probs = self.call_duplex_read_mod_probs(duplex_read)

        template_mm, template_ml = format_mm_ml_tags(
            seq=duplex_read_probs["read_sequence"],
            poss=duplex_read_probs["template_positions"],
            probs=duplex_read_probs["template_probs"],
            mod_bases=self.model_metadata["mod_bases"],
            can_base=self.model_metadata["can_base"],
            strand="+",
        )
        complement_mm, complement_ml = format_mm_ml_tags(
            seq=duplex_read_probs["read_sequence"],
            poss=duplex_read_probs["complement_positions"],
            probs=duplex_read_probs["complement_probs"],
            mod_bases=self.model_metadata["mod_bases"],
            can_base=revcomp(self.model_metadata["can_base"]),
            strand="-",
        )
        mm_tag = f"MM:Z:{template_mm + complement_mm}"
        ml_tag_values = template_ml + complement_ml
        ml_tag_stringed = ",".join([str(x) for x in ml_tag_values])
        ml_tag = f"ML:B:C,{ml_tag_stringed}"

        return mm_tag, ml_tag


################
# POD5+BAM CLI #
################


def mods_tags_to_str(mods_tags):
    # TODO these operations are often quite slow
    return [
        f"MM:Z:{mods_tags[0]}",
        f"ML:B:C,{','.join(map(str, mods_tags[1]))}",
    ]


def prepare_batches(read_errs, model_metadata, batch_size, ref_anchored):
    out_read_errs = []
    for io_read, err in read_errs:
        if io_read is None:
            out_read_errs.append((None, None, err))
            continue
        try:
            remora_read = io_read.into_remora_read(ref_anchored)
        except RemoraError as e:
            LOGGER.debug(f"Remora read prep error: {e}")
            out_read_errs.append((None, None, "Remora read prep error"))
            continue
        motifs = [Motif(*mot) for mot in model_metadata["motifs"]]
        remora_read.set_motif_focus_bases(motifs)
        remora_read.prepare_batches(model_metadata, batch_size)
        if len(remora_read.batches) == 0:
            out_read_errs.append((None, None, "No mod calls"))
            continue
        out_read_errs.append((io_read, remora_read, None))
    return out_read_errs


if _PROF_PREP_FN:
    _prepare_batches_wrapper = prepare_batches

    def prepare_batches(*args, **kwargs):
        import cProfile

        prof = cProfile.Profile()
        retval = prof.runcall(_prepare_batches_wrapper, *args, **kwargs)
        prof.dump_stats(_PROF_PREP_FN)
        return retval


def run_model(read_errs, model, model_metadata, ref_anchored):
    out_read_errs = []
    for io_read, remora_read, err in read_errs:
        if err is not None:
            out_read_errs.append((None, err))
            continue
        # TODO super-batch reads to optimize performance for short reads
        nn_out, labels, pos = remora_read.run_model(model)
        probs = softmax_axis1(nn_out)[:, 1:].astype(np.float64)
        mod_tags = mods_tags_to_str(
            format_mm_ml_tags(
                seq=remora_read.str_seq,
                poss=pos,
                probs=probs,
                mod_bases=model_metadata["mod_bases"],
                can_base=model_metadata["can_base"],
            )
        )
        io_read.full_align["tags"] = [
            tag
            for tag in io_read.full_align["tags"]
            if not (tag.startswith("MM") or tag.startswith("ML"))
        ]
        io_read.full_align["tags"].extend(mod_tags)
        if ref_anchored:
            io_read.full_align["cigar"] = f"{len(io_read.ref_seq)}M"
            io_read.full_align["seq"] = (
                io_read.ref_seq
                if io_read.ref_reg.strand == "+"
                else revcomp(io_read.ref_seq)
            )
            io_read.full_align["qual"] = "*"
        out_read_errs.append((io_read, None))
    return out_read_errs


if _PROF_MODEL_FN:
    _run_model_wrapper = run_model

    def run_model(*args, **kwargs):
        import cProfile

        prof = cProfile.Profile()
        retval = prof.runcall(_run_model_wrapper, *args, **kwargs)
        prof.dump_stats(_PROF_MODEL_FN)
        return retval


def infer_from_pod5_and_bam(
    pod5_path,
    in_bam_path,
    model,
    model_metadata,
    out_bam_path,
    num_reads=None,
    queue_max=1_000,
    num_extract_alignment_workers=1,
    num_prep_batch_workers=1,
    batch_size=constants.DEFAULT_BATCH_SIZE,
    skip_non_primary=True,
    ref_anchored=False,
):
    bam_idx = ReadIndexedBam(in_bam_path, skip_non_primary, req_tags={"mv"})
    with pod5.Reader(Path(pod5_path)) as pod5_fh:
        read_ids, num_reads = get_read_ids(bam_idx, pod5_fh, num_reads)

    signals = BackgroundIter(
        iter_signal,
        args=(pod5_path,),
        kwargs={
            "num_reads": num_reads,
            "read_ids": read_ids,
            "rev_sig": model_metadata["reverse_signal"],
        },
        name="ExtractSignal",
        use_process=True,
        q_maxsize=queue_max,
    )
    reads = MultitaskMap(
        extract_alignments,
        signals,
        num_workers=num_extract_alignment_workers,
        args=(bam_idx, model_metadata["reverse_signal"]),
        name="AddAlignments",
        use_process=True,
        q_maxsize=queue_max,
    )
    reads = MultitaskMap(
        prepare_batches,
        reads,
        num_workers=num_prep_batch_workers,
        args=(model_metadata, batch_size, ref_anchored),
        name="PrepBatches",
        use_process=True,
        q_maxsize=queue_max,
    )

    errs = defaultdict(int)
    pysam_save = pysam.set_verbosity(0)
    sig_called = 0
    in_bam = out_bam = pbar = None
    try:
        in_bam = pysam.AlignmentFile(in_bam_path, "rb")
        out_bam = pysam.AlignmentFile(out_bam_path, "wb", template=in_bam)
        pbar = tqdm(
            smoothing=0,
            total=num_reads,
            unit=" Reads",
            desc="Inferring mods",
            disable=os.environ.get("LOG_SAFE", False),
        )
        # TODO add a batch -> run_model -> unbatch function here to run model
        # more efficiently on larger batch size
        for read_errs in reads:
            pbar.update()
            if len(read_errs) == 0:
                errs["No valid mappings"] += 1
                continue

            sig_called += sum(
                read.dacs.size for read, _, _ in read_errs if read is not None
            )
            msps = sig_called / 1_000_000 / pbar.format_dict["elapsed"]
            pbar.set_postfix_str(f"{msps:.2f} Msamps/s", refresh=False)
            read_errs = run_model(
                read_errs, model, model_metadata, ref_anchored
            )

            for io_read, err in read_errs:
                if io_read is None:
                    errs[err] += 1
                    continue
                out_bam.write(
                    pysam.AlignedSegment.from_dict(
                        io_read.full_align, out_bam.header
                    )
                )
    finally:
        if pbar is not None:
            pbar.close()
        if in_bam is not None:
            in_bam.close()
        if out_bam is not None:
            out_bam.close()
    pysam.set_verbosity(pysam_save)

    if len(errs) > 0:
        err_types = sorted([(num, err) for err, num in errs.items()])[::-1]
        err_str = "\n".join(f"{num:>7} : {err:<80}" for num, err in err_types)
        LOGGER.info(f"Unsuccessful read reasons:\n{err_str}")


if _PROF_MAIN_FN:
    _infer_from_pod5_and_bam_wrapper = infer_from_pod5_and_bam

    def infer_from_pod5_and_bam(*args, **kwargs):
        import cProfile

        prof = cProfile.Profile()
        retval = prof.runcall(_infer_from_pod5_and_bam_wrapper, *args, **kwargs)
        prof.dump_stats(_PROF_MAIN_FN)
        return retval


def check_simplex_alignments(*, simplex_index, duplex_index, pairs):
    """Check that valid pairs are found

    Args:
        simplex_index (ReadIndexedBam): Simplex basecalls bam index
        duplex_index (ReadIndexedBam): Duplex basecalls bam index
        pairs (list): List of read pair strings
    """
    if len(pairs) == 0:
        raise ValueError("no pairs found in file")
    all_paired_read_ids = set(itertools.chain(*pairs))
    simplex_read_ids = set(simplex_index.read_ids)
    duplex_read_ids = set(duplex_index.read_ids)
    count_paired_simplex_alignments = sum(
        [1 for read_id in all_paired_read_ids if read_id in simplex_read_ids]
    )
    if count_paired_simplex_alignments == 0:
        raise ValueError("zero simplex alignments found")

    # valid meaning we have all the parts to perform inference
    valid_read_pairs = [
        (t, c)
        for t, c in pairs
        if all(read_id in simplex_read_ids for read_id in (t, c))
        and t in duplex_read_ids
    ]
    return valid_read_pairs, len(valid_read_pairs)


def prep_duplex_read_builder(simplex_index, pod5_path):
    """Prepare a duplex pair builder object. For example pass to MultitaskMap
    prep_func argument.

    Args:
        simplex_index (io.ReadIndexedBam): Read indexed BAM file handle
        pod5_path (pod5.reader.Reader): POD5 file handle
    """
    builder = DuplexPairsBuilder(
        simplex_index=simplex_index,
        pod5_path=pod5_path,
    )
    return [builder], {}


def iter_duplexed_io_reads(read_id_pair, builder):
    return builder.make_read_pair(read_id_pair)


def infer_duplex(
    *,
    simplex_pod5_path: str,
    simplex_bam_path: str,
    duplex_bam_path: str,
    pairs_path: str,
    model,
    model_metadata,
    out_bam,
    num_extract_alignment_threads,
    num_duplex_prep_workers,
    num_infer_threads,
    num_reads=None,
    skip_non_primary=True,
    duplex_deliminator=";",
):
    duplex_bam_index = ReadIndexedBam(
        duplex_bam_path,
        skip_non_primary=skip_non_primary,
        req_tags=set(),
        read_id_converter=lambda k: k.split(duplex_deliminator)[0],
    )

    simplex_bam_index = ReadIndexedBam(
        simplex_bam_path, skip_non_primary=True, req_tags={"mv"}
    )
    with open(pairs_path, "r") as fh:
        pairs = [tuple(line.split()) for line in fh]
    valid_pairs, num_valid_reads = check_simplex_alignments(
        simplex_index=simplex_bam_index,
        duplex_index=duplex_bam_index,
        pairs=pairs,
    )
    if num_reads is not None:
        num_reads = min(num_valid_reads, num_reads)

    # source of pipeline, template, complement read ID pairs
    # produces template, complement read id tuples
    def iter_pairs(pairs, num_reads):
        for pair_num, pair in enumerate(pairs):
            if num_reads is not None and pair_num >= num_reads:
                return
            yield pair

    read_id_pairs = BackgroundIter(
        iter_pairs,
        kwargs=dict(pairs=valid_pairs, num_reads=num_reads),
        use_process=True,
    )

    # consumes: tuple of template, complement read Ids
    # prep: open resources for Pod5 and simplex BAM
    # produces: (io.Read, io.Read), str
    io_read_pairs_results = MultitaskMap(
        iter_duplexed_io_reads,
        read_id_pairs,
        prep_func=prep_duplex_read_builder,
        args=(simplex_bam_index, simplex_pod5_path),
        name="BuildDuplexedIoReads",
        num_workers=num_extract_alignment_threads,
        use_process=True,
    )

    def make_duplex_reads(read_pair_result, duplex_index, bam_file_handle):
        read_pair, err = read_pair_result
        if err is not None or read_pair is None:
            return None, err
        template, complement = read_pair
        if template.read_id not in duplex_index:
            return None, "duplex BAM record not found for read_id"
        for pointer in duplex_index[template.read_id]:
            bam_file_handle.seek(pointer)
            bam_record = next(bam_file_handle)
            duplex_read = DuplexRead.from_reads_and_alignment(
                template_read=template,
                complement_read=complement,
                duplex_alignment=bam_record,
            )

            return duplex_read, None

    # consumes: tuple of io.Reads (template, complement)
    # produces: (DuplexRead, str), for inference by the model
    duplex_aln = pysam.AlignmentFile(duplex_bam_path, "rb", check_sq=False)
    duplex_reads = MultitaskMap(
        make_duplex_reads,
        io_read_pairs_results,
        num_workers=num_duplex_prep_workers,
        args=(duplex_bam_index, duplex_aln),
        name="MakeDuplexReads",
        use_process=True,
    )

    def add_mod_mappings_to_alignment(duplex_read_result, caller):
        duplex_read, err = duplex_read_result
        if err is not None:
            return None, err
        mod_tags = caller.call_duplex_read_mods(duplex_read)
        mod_tags = list(mod_tags)
        assert len(mod_tags) == 2
        record = duplex_read.duplex_alignment
        record = copy(record)
        record["tags"] = [
            tag
            for tag in record["tags"]
            if not (tag.startswith("MM") or tag.startswith("ML"))
        ]
        record["tags"].extend(mod_tags)
        return record, None

    duplex_caller = DuplexReadModCaller(model, model_metadata)
    use_process = True
    if isinstance(model, RecursiveScriptModule):
        use_process = next(model.parameters()).device.type == "cpu"

    # consumes: Result[DuplexReads, str]
    # produces: (dict, str) the dict is the BAM record with mods
    # added/substituted mod tags
    alignment_records_with_mod_tags = MultitaskMap(
        add_mod_mappings_to_alignment,
        duplex_reads,
        num_workers=num_infer_threads,
        args=(duplex_caller,),
        name="InferMods",
        use_process=use_process,
    )

    errs = defaultdict(int)
    pysam_save = pysam.set_verbosity(0)
    with pysam.AlignmentFile(duplex_bam_path, "rb", check_sq=False) as in_bam:
        with pysam.AlignmentFile(out_bam, "wb", template=in_bam) as out_bam:
            for mod_read_mapping, err in tqdm(
                alignment_records_with_mod_tags,
                total=num_reads,
                disable=os.environ.get("LOG_SAFE", False),
            ):
                if err is not None:
                    errs[err] += 1
                    continue
                out_bam.write(
                    pysam.AlignedSegment.from_dict(
                        mod_read_mapping, out_bam.header
                    )
                )
    pysam.set_verbosity(pysam_save)
    duplex_aln.close()

    if len(errs) > 0:
        err_types = sorted([(num, err) for err, num in errs.items()])[::-1]
        err_str = "\n".join(f"{num:>7} : {err:<80}" for num, err in err_types)
        LOGGER.info(f"Unsuccessful read reasons:\n{err_str}")


if __name__ == "__main__":
    NotImplementedError("This is a module.")



import os
from pathlib import Path
from collections import defaultdict

import pod5
import numpy as np
from tqdm import tqdm

from remora import log, RemoraError
from remora.util import MultitaskMap, BackgroundIter, get_read_ids
from remora.io import ReadIndexedBam, iter_signal, extract_alignments
from remora.data_chunks import RemoraRead, RemoraDataset, compute_ref_to_signal

LOGGER = log.get_logger()


####################
# Chunk extraction #
####################


def extract_chunks(
    read_errs,
    int_label,
    motifs,
    focus_ref_pos,
    sig_map_refiner,
    max_chunks_per_read,
    chunk_context,
    kmer_context_bases,
    base_pred,
    base_start_justify,
    offset,
    basecall_anchor,
):
    read_chunks = []
    for read_idx, (io_read, err) in enumerate(read_errs):
        if io_read is None:
            read_chunks.append(tuple((io_read, err)))
            continue
        if io_read.ref_seq is None:
            read_chunks.append(
                tuple((None, "No reference sequence (missing MD tag)"))
            )
            continue
        if basecall_anchor:
            remora_read = io_read.into_remora_read(use_reference_anchor=False)
            remora_read.focus_bases = io_read.get_basecall_anchored_focus_bases(
                motifs=motifs,
                select_focus_reference_positions=focus_ref_pos,
            )
            remora_read.labels = np.full(len(io_read.seq), int_label, dtype=int)
        else:
            io_read.ref_to_signal = compute_ref_to_signal(
                io_read.query_to_signal,
                io_read.cigar,
            )
            assert io_read.ref_to_signal.size == len(io_read.ref_seq) + 1, (
                "discordant ref seq lengths: move+cigar:"
                f"{io_read.ref_to_signal.size} ref_seq:{len(io_read.ref_seq)}"
            )
            trim_dacs = io_read.dacs[
                io_read.ref_to_signal[0] : io_read.ref_to_signal[-1]
            ]
            shift_ref_to_sig = io_read.ref_to_signal - io_read.ref_to_signal[0]
            remora_read = RemoraRead(
                dacs=trim_dacs,
                shift=io_read.shift_dacs_to_norm,
                scale=io_read.scale_dacs_to_norm,
                seq_to_sig_map=shift_ref_to_sig,
                str_seq=io_read.ref_seq,
                labels=np.full(len(io_read.ref_seq), int_label, dtype=int),
                read_id=io_read.read_id,
            )
            if focus_ref_pos is not None:
                # todo(arand) make a test that exercises this code path
                remora_read.focus_bases = io_read.get_filtered_focus_positions(
                    focus_ref_pos
                )
            else:
                remora_read.set_motif_focus_bases(motifs)

        remora_read.refine_signal_mapping(sig_map_refiner)
        remora_read.downsample_focus_bases(max_chunks_per_read)
        try:
            remora_read.check()
        except RemoraError as e:
            LOGGER.debug(f"Read prep failed: {e}")
            continue
        read_align_chunks = list(
            remora_read.iter_chunks(
                chunk_context,
                kmer_context_bases,
                base_pred,
                base_start_justify,
                offset,
                check_chunks=True,
            )
        )
        LOGGER.debug(
            f"extracted {len(read_align_chunks)} chunks from {io_read.read_id} "
            f"alignment {read_idx}"
        )
        read_chunks.append(tuple((read_align_chunks, None)))

    return read_chunks


############
# Pipeline #
############


def extract_chunk_dataset(
    bam_fn,
    pod5_path,
    out_fn,
    mod_base,
    mod_base_control,
    motifs,
    focus_ref_pos,
    chunk_context,
    min_samps_per_base,
    max_chunks_per_read,
    sig_map_refiner,
    base_pred,
    kmer_context_bases,
    base_start_justify,
    offset,
    num_reads,
    num_extract_alignment_threads,
    num_extract_chunks_threads,
    skip_non_primary=True,
    basecall_anchor=False,
    rev_sig=False,
):
    bam_idx = ReadIndexedBam(bam_fn, skip_non_primary)
    with pod5.Reader(Path(pod5_path)) as pod5_fh:
        read_ids, num_reads = get_read_ids(bam_idx, pod5_fh, num_reads)
    if num_reads == 0:
        return

    LOGGER.info(
        f"Making {'basecall' if basecall_anchor else 'reference'}-"
        f"anchored training data"
    )
    LOGGER.info("Allocating memory for output tensors")
    # initialize empty dataset with pre-allocated memory
    dataset = RemoraDataset.allocate_empty_chunks(
        num_chunks=max_chunks_per_read * num_reads,
        chunk_context=chunk_context,
        kmer_context_bases=kmer_context_bases,
        min_samps_per_base=min_samps_per_base,
        base_pred=base_pred,
        mod_bases=[] if mod_base_control else [mod_base[0]],
        mod_long_names=[] if mod_base_control else [mod_base[1]],
        motifs=[mot.to_tuple() for mot in motifs],
        reverse_signal=rev_sig,
        sig_map_refiner=sig_map_refiner,
        base_start_justify=base_start_justify,
        offset=offset,
    )

    LOGGER.info("Processing reads")
    signals = BackgroundIter(
        iter_signal,
        args=(pod5_path,),
        kwargs={
            "num_reads": num_reads,
            "read_ids": read_ids,
            "rev_sig": rev_sig,
        },
        name="ExtractSignal",
        use_process=True,
    )
    reads = MultitaskMap(
        extract_alignments,
        signals,
        num_workers=num_extract_alignment_threads,
        args=(bam_idx, rev_sig),
        name="AddAlignments",
        use_process=True,
    )
    chunks = MultitaskMap(
        extract_chunks,
        reads,
        num_workers=num_extract_chunks_threads,
        args=[
            0 if mod_base_control else 1,
            motifs,
            focus_ref_pos,
            sig_map_refiner,
            max_chunks_per_read,
            chunk_context,
            kmer_context_bases,
            base_pred,
            base_start_justify,
            offset,
            basecall_anchor,
        ],
        name="ExtractChunks",
        use_process=True,
    )

    errs = defaultdict(int)
    for read_chunks in tqdm(
        chunks,
        total=num_reads,
        smoothing=0,
        unit=" Reads",
        desc="Extracting chunks",
        disable=os.environ.get("LOG_SAFE", False),
    ):
        if len(read_chunks) == 0:
            errs["No chunks extracted"] += 1
            continue
        for read_align_chunks, err in read_chunks:
            if read_align_chunks is None:
                errs[err] += 1
                continue
            for chunk in read_align_chunks:
                try:
                    dataset.add_chunk(chunk)
                except RemoraError as e:
                    errs[str(e)] += 1

    if len(errs) > 0:
        err_types = sorted([(num, err) for err, num in errs.items()])[::-1]
        err_str = "\n".join(f"{num:>7} : {err:<80}" for num, err in err_types)
        LOGGER.info(f"Unsuccessful read/chunk reasons:\n{err_str}")

    dataset.clip_chunks()
    dataset.shuffle()
    dataset.save(out_fn)

    LOGGER.info(f"Extracted {dataset.nchunks} chunks from {num_reads} reads.")
    LOGGER.info(f"Label distribution: {dataset.get_label_counts()}")



import numpy as np

DEFAULT_START_TRIM = 1
DEFAULT_END_TRIM = 1


def clip_sig(sig, seq_to_sig):
    sig = sig[seq_to_sig[0] : seq_to_sig[-1]]
    return sig, seq_to_sig - seq_to_sig[0]


def cumsum0(sig):
    sig_cs = np.empty(sig.size + 1)
    sig_cs[0] = 0
    sig_cs[1:] = np.cumsum(sig)
    return sig_cs


def compute_cumsum_and_mean(sig, seq_to_sig, dwells):
    sig_cs = cumsum0(sig)
    with np.errstate(divide="ignore", invalid="ignore"):
        base_means = np.diff(sig_cs[seq_to_sig]) / dwells
        base_means[np.isinf(base_means)] = np.nan
    return base_means


def compute_trim_cumsum_and_mean(
    sig,
    seq_to_sig,
    dwells,
    st_trim=DEFAULT_START_TRIM,
    en_trim=DEFAULT_END_TRIM,
):
    sig_cs = cumsum0(sig)
    trim_sts = np.minimum(sig.size, seq_to_sig[:-1] + st_trim)
    trim_ens = np.maximum(0, seq_to_sig[1:] - en_trim)
    trim_sums = sig_cs[trim_ens] - sig_cs[trim_sts]
    trim_dwells = np.maximum(0, dwells - st_trim - en_trim)
    with np.errstate(divide="ignore", invalid="ignore"):
        trim_means = trim_sums / trim_dwells
        trim_means[np.isinf(trim_means)] = np.nan
    return trim_means


def compute_dwell(sig, seq_to_sig, **kwargs):
    return {"dwell": np.diff(seq_to_sig).astype(np.float32)}


def compute_dwell_mean(sig, seq_to_sig, **kwargs):
    dwells = compute_dwell(sig, seq_to_sig, **kwargs)["dwell"]
    sig, seq_to_sig = clip_sig(sig, seq_to_sig)
    base_means = compute_cumsum_and_mean(sig, seq_to_sig, dwells)
    return {"dwell": dwells, "mean": base_means}


def compute_dwell_mean_sd(sig, seq_to_sig, **kwargs):
    dwells = compute_dwell(sig, seq_to_sig, **kwargs)["dwell"]
    sig, seq_to_sig = clip_sig(sig, seq_to_sig)
    base_means = compute_cumsum_and_mean(sig, seq_to_sig, dwells)
    ss_cs = cumsum0(np.square(sig))
    with np.errstate(divide="ignore", invalid="ignore"):
        base_sds = np.sqrt(
            np.maximum(
                np.diff(ss_cs[seq_to_sig]) / dwells - np.square(base_means),
                0,
            )
        )
        base_sds[np.isinf(base_sds)] = np.nan
    return {"dwell": dwells, "mean": base_means, "sd": base_sds}


def compute_trimmean(sig, seq_to_sig, **kwargs):
    st_trim = kwargs.get("start_trim", DEFAULT_START_TRIM)
    en_trim = kwargs.get("end_trim", DEFAULT_END_TRIM)

    dwells = compute_dwell(sig, seq_to_sig)["dwell"]
    sig, seq_to_sig = clip_sig(sig, seq_to_sig)
    trim_means = compute_trim_cumsum_and_mean(
        sig, seq_to_sig, dwells, st_trim, en_trim
    )
    return {"dwells": dwells, "trimmean": trim_means}


def compute_trimmean_trimsd(sig, seq_to_sig, **kwargs):
    st_trim = kwargs.get("start_trim", DEFAULT_START_TRIM)
    en_trim = kwargs.get("end_trim", DEFAULT_END_TRIM)

    dwells = compute_dwell(sig, seq_to_sig)["dwell"]
    sig, seq_to_sig = clip_sig(sig, seq_to_sig)
    trim_means = compute_trim_cumsum_and_mean(
        sig, seq_to_sig, dwells, st_trim, en_trim
    )
    ss_cs = cumsum0(np.square(sig))
    trim_dwells = np.maximum(0, dwells - st_trim - en_trim)

    trim_sts = np.minimum(sig.size, seq_to_sig[:-1] + st_trim)
    trim_ens = np.maximum(0, seq_to_sig[1:] - en_trim)
    trim_ss_sums = ss_cs[trim_ens] - ss_cs[trim_sts]
    trim_dwells = np.maximum(0, dwells - st_trim - en_trim)
    with np.errstate(divide="ignore", invalid="ignore"):
        trim_sds = np.sqrt(
            np.maximum(
                0,
                (trim_ss_sums / trim_dwells) - np.square(trim_means),
            )
        )
        trim_sds[np.isinf(trim_sds)] = np.nan
    return {"dwell": dwells, "trimmean": trim_means, "trimsd": trim_sds}


METRIC_FUNCS = {
    "dwell": compute_dwell,
    "dwell_mean": compute_dwell_mean,
    "dwell_mean_sd": compute_dwell_mean_sd,
    "dwell_trimmean": compute_trimmean,
    "dwell_trimmean_trimsd": compute_trimmean_trimsd,
}



DEFAULT_NN_SIZE = 64
DEFAULT_BATCH_SIZE = 1024
DEFAULT_CHUNK_CONTEXT = (50, 50)
DEFAULT_MIN_SAMPLES_PER_BASE = 5
DEFAULT_KMER_CONTEXT_BASES = (4, 4)
DEFAULT_KMER_LEN = sum(DEFAULT_KMER_CONTEXT_BASES) + 1
DEFAULT_FOCUS_OFFSET = 100
DEFAULT_VAL_PROP = 0.01
DEFAULT_EPOCHS = 50
DEFAULT_WEIGHT_DECAY = 1e-4
DEFAULT_FILT_FRAC = 0.1
DEFAULT_LR = 0.001

DEFAULT_SCHEDULER = "StepLR"
DEFAULT_SCH_VALUES = {"step_size": 10, "gamma": 0.6}
TYPE_CONVERTERS = {"str": str, "int": int, "float": float}

SGD_OPT = "sgd"
ADAM_OPT = "adam"
ADAMW_OPT = "adamw"
OPTIMIZERS = (ADAMW_OPT, SGD_OPT, ADAM_OPT)

FINAL_MODEL_FILENAME = "model_final.checkpoint"
FINAL_TORCHSCRIPT_MODEL_FILENAME = "model_final.pt"
SAVE_DATASET_FILENAME = "remora_train_data.npz"

BEST_MODEL_FILENAME = "model_best.checkpoint"
BEST_TORCHSCRIPT_MODEL_FILENAME = "model_best.pt"

MODEL_VERSION = 3

DEFAULT_REFINE_HBW = 5

MODBASE_MODEL_NAME = "modbase_model.pt"
MODEL_DATA_DIR_NAME = "trained_models"

"""
The default model is the first key at every level after the pore and mod.
E.g. for "dna_r10.4.1_e8.2_400bps" and "5mc" the default model is
CG_sup_v3.5.1_2.
"""

# R9 5mC CG-context models
_R9_5mc_CG_models = {
    "sup": {"v3.5.1": {0: "dna_r9.4.1_e8_sup_v3.5.1_5mc_CG_v0"}},
    "hac": {"v3.5.1": {0: "dna_r9.4.1_e8_hac_v3.5.1_5mc_CG_v0"}},
    "fast": {"v3.5.1": {0: "dna_r9.4.1_e8_fast_v3.5.1_5mc_CG_v0"}},
}

# kit14 400bps 5mC CG-context models (contains 5kHz and 4kHz models)
_kit14_5mc_CG_models = {
    "sup": {
        "v4.2.0": {2: "dna_r10.4.1_e8.2_5khz_400bps_sup_v4.2.0_5mc_CG_v2"},
        "v4.1.0": {2: "dna_r10.4.1_e8.2_4khz_400bps_sup_v4.1.0_5mc_CG_v2"},
        "v3.5.1": {2: "dna_r10.4.1_e8.2_400bps_sup_v3.5.1_5mc_CG_v2"},
    },
    "hac": {
        "v4.2.0": {2: "dna_r10.4.1_e8.2_5khz_400bps_hac_v4.2.0_5mc_CG_v2"},
        "v4.1.0": {2: "dna_r10.4.1_e8.2_4khz_400bps_hac_v4.1.0_5mc_CG_v2"},
        "v3.5.1": {2: "dna_r10.4.1_e8.2_400bps_hac_v3.5.1_5mc_CG_v2"},
    },
    "fast": {
        "v4.2.0": {2: "dna_r10.4.1_e8.2_5khz_400bps_fast_v4.2.0_5mc_CG_v2"},
        "v4.1.0": {2: "dna_r10.4.1_e8.2_4khz_400bps_fast_v4.1.0_5mc_CG_v2"},
        "v3.5.1": {2: "dna_r10.4.1_e8.2_400bps_fast_v3.5.1_5mc_CG_v2"},
    },
}

# kit14 400bps 5hmc_5mC CG-context models (contains 5kHz and 4kHz models)
_kit14_5hmc_5mc_CG_models = {
    "sup": {
        "v4.2.0": {2: "dna_r10.4.1_e8.2_5khz_400bps_sup_v4.2.0_5hmc_5mc_CG_v2"},
        "v4.1.0": {2: "dna_r10.4.1_e8.2_4khz_400bps_sup_v4.1.0_5hmc_5mc_CG_v2"},
        "v4.0.0": {2: "dna_r10.4.1_e8.2_400bps_sup_v4.0.0_5hmc_5mc_CG_v2"},
    },
    "hac": {
        "v4.2.0": {2: "dna_r10.4.1_e8.2_5khz_400bps_hac_v4.2.0_5hmc_5mc_CG_v2"},
        "v4.1.0": {2: "dna_r10.4.1_e8.2_4khz_400bps_hac_v4.1.0_5hmc_5mc_CG_v2"},
        "v4.0.0": {2: "dna_r10.4.1_e8.2_400bps_hac_v4.0.0_5hmc_5mc_CG_v2"},
    },
    "fast": {
        "v4.2.0": {
            2: "dna_r10.4.1_e8.2_5khz_400bps_fast_v4.2.0_5hmc_5mc_CG_v2"
        },
        "v4.1.0": {
            2: "dna_r10.4.1_e8.2_4khz_400bps_fast_v4.1.0_5hmc_5mc_CG_v2"
        },
        "v4.0.0": {2: "dna_r10.4.1_e8.2_400bps_fast_v4.0.0_5hmc_5mc_CG_v2"},
    },
}

# kit14 260bps 5hmC_5mC CG-context models
_kit14_260bps_5hmc_5mc_CG_models = {
    "sup": {
        "v4.0.0": {2: "dna_r10.4.1_e8.2_260bps_sup_v4.0.0_5hmc_5mc_CG_v2"},
    },
    "hac": {
        "v4.0.0": {2: "dna_r10.4.1_e8.2_260bps_hac_v4.0.0_5hmc_5mc_CG_v2"},
    },
    "fast": {
        "v4.0.0": {2: "dna_r10.4.1_e8.2_260bps_fast_v4.0.0_5hmc_5mc_CG_v2"},
    },
}

# all-context models (contains 5kHz and 4kHz models)
_kit14_5mc_ac_models = {
    "sup": {
        "v4.2.0": {2: "dna_r10.4.1_e8.2_5khz_400bps_sup_v4.2.0_5mc_v2"},
        "v4.0.1": {2: "res_dna_r10.4.1_e8.2_4khz_400bps_sup_v4.0.1_5mc_v2"},
    },
}
_kit14_6ma_ac_models = {
    "sup": {
        "v4.2.0": {2: "dna_r10.4.1_e8.2_5khz_400bps_sup_v4.2.0_6ma_v2"},
        "v4.0.1": {2: "res_dna_r10.4.1_e8.2_4khz_400bps_sup_v4.0.1_6mA_v2"},
    },
}

MODEL_DICT = {
    "dna_r9.4.1_e8": {
        "5mc": {
            "CG": _R9_5mc_CG_models,
        },
    },
    "dna_r10.4.1_e8.2_400bps": {
        "5mc": {
            "C": _kit14_5mc_ac_models,
            "CG": _kit14_5mc_CG_models,
        },
        "6ma": {
            "A": _kit14_6ma_ac_models,
        },
        "5hmc_5mc": {
            "CG": _kit14_5hmc_5mc_CG_models,
        },
    },
    "dna_r10.4.1_e8.2_260bps": {
        "5hmc_5mc": {
            "CG": _kit14_260bps_5hmc_5mc_CG_models,
        },
    },
}

DEFAULT_REFINE_SHORT_DWELL_PARAMS = (4, 3, 0.5)
REFINE_ALGO_VIT_NAME = "Viterbi"
REFINE_ALGO_DWELL_PEN_NAME = "dwell_penalty"
REFINE_ALGOS = (REFINE_ALGO_DWELL_PEN_NAME, REFINE_ALGO_VIT_NAME)
DEFAULT_REFINE_ALGO = REFINE_ALGO_DWELL_PEN_NAME

PA_TO_NORM_SCALING_FACTOR = 1.4826

DEFAULT_PLOT_FIG_SIZE = (40, 10)



# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False

import numpy as np

from remora import RemoraError
from remora.constants import REFINE_ALGO_VIT_NAME, REFINE_ALGO_DWELL_PEN_NAME

cimport cython
from libc.math cimport HUGE_VALF

cdef float LARGE_SCORE = 100

# function pointer for core by-base scoring functions
ctypedef void (*core_func_ptr)(
    float[::1],        # curr_scores
    int[::1],          # curr_tb
    const float[::1],  # prev_scores
    const float,       # curr_level
    const float[::1],  # curr_signal
    int,               # band_start_diff
    const float[::1],  # dwell_penalty
)


###########
# Banding #
###########

def adjust_seq_band(int[:, ::1] seq_band, int min_step=2):
    """Adjust band boundaries to disallow invalid paths. Namely, make sure each
    band start and end is at least one greater than the previous.

    Note input band will be updated in place

    Args:
        seq_band (np.ndarray): int32 np.ndarray with shape
            (2, seq_len = levels.shape[0]). The first row contains the lower
            band boundaries in signal coordinates and the second row contains
            the upper boundaries in signal coordinates. The first lower bound
            should be 0 and last upper bound should be signal.shape[0].
        min_step (int): Minimum step between one base and the next to enforce
            in band adjustment.
    """
    cdef int band_min = seq_band[0, 0]
    cdef int seq_pos
    # fix starts to make sure each start is at least min_step less than prev
    for seq_pos in range(seq_band.shape[1] - 2, -1, -1):
        if seq_band[0, seq_pos] > seq_band[0, seq_pos + 1] - min_step:
            seq_band[0, seq_pos] = seq_band[0, seq_pos + 1] - min_step
    # proceed through beginning of band ensuring only valid positions
    seq_band[0, 0] = band_min
    seq_pos = 1
    while seq_band[0, seq_pos] <= seq_band[0, seq_pos - 1]:
        seq_band[0, seq_pos] = seq_band[0, seq_pos - 1] + 1
        seq_pos += 1

    cdef int band_max = seq_band[1, seq_band.shape[1] - 1]
    # fix ends to make sure each end is at least min_step more than prev
    for seq_pos in range(1, seq_band.shape[1]):
        if seq_band[1, seq_pos] < seq_band[1, seq_pos - 1] + min_step:
            seq_band[1, seq_pos] = seq_band[1, seq_pos - 1] + min_step
    # proceed through end of band ensuring only valid positions
    seq_band[1, seq_band.shape[1] - 1] = band_max
    seq_pos = seq_band.shape[1] - 2
    while seq_band[1, seq_pos] >= seq_band[1, seq_pos + 1]:
        seq_band[1, seq_pos] = seq_band[1, seq_pos + 1] - 1
        seq_pos -= 1


####################
# Level Extraction #
####################

cdef inline int index_from_int_kmer(
    const int[::1] int_kmer,
    int kmer_len
):
    cdef int idx = 0
    cdef int kmer_pos
    for kmer_pos in range(kmer_len):
        idx += int_kmer[kmer_len - kmer_pos - 1] * (4 ** kmer_pos)
    return idx


def extract_levels(
    const int[::1] int_seq,
    const float[::1] int_kmer_levels,
    int kmer_len,
    int center_idx
):
    levels = np.zeros(int_seq.shape[0], dtype=np.float32)
    cdef float[::1] levels_mv = levels
    cdef int pos, center_pos, kmer_idx
    for pos in range(int_seq.shape[0] - kmer_len + 1):
        center_pos = pos + center_idx
        kmer_idx = index_from_int_kmer(int_seq[pos : pos + kmer_len], kmer_len)
        levels_mv[center_pos] = int_kmer_levels[kmer_idx]
    return levels


######################
# Mapping Refinement #
######################

cdef inline float score(float s, float l):
    """Find squared difference between sample and level.

    Args:
        s (float): sample
        l (float): level
    """
    cdef float tmp = s - l
    return tmp * tmp


cdef void banded_traceback(
    int[::1] path,
    const int[:, ::1] seq_band,
    const int[::1] base_offsets,
    const int[::1] traceback,
):
    """Perform traceback to determine path from a forward pass.

    Args:
        path (1D int array): To be populated by this function. Path will
            contain start position for each base within the signal array.
        seq_band (const 2D int array): Contains band boundaries for each base in
            signal coordinates.
        base_offsets (const 1D int array): Offset to the beginning of each base
            within the ragged traceback array.
        traceback (const 1D int array): Ragged array representing each position
            within the band. Each element contains the number of signal points
            backwards until the first point assigned to this base.
    """
    cdef int sig_lookup_pos, next_sig_offset, base_idx
    # set start to 0 and end to sig_len
    path[0] = 0
    path[path.shape[0] - 1] = seq_band[1, seq_band.shape[1] - 1]
    for base_idx in range(path.shape[0] - 2, 0, -1):
        # signal position to lookup for this traceback step
        sig_lookup_pos = path[base_idx + 1] - 1
        next_sig_offset = traceback[
            base_offsets[base_idx] + sig_lookup_pos - seq_band[0, base_idx]
        ]
        # record position in which base_idx starts
        path[base_idx] = sig_lookup_pos - next_sig_offset

cdef void banded_forward_dwell_penalty_step(
    float[::1] curr_scores,
    int[::1] curr_tb,
    float[::1] prev_scores,
    const float curr_level,
    const float[::1] curr_signal,
    int band_start_diff,
    const float[::1] dwell_penalty,
):
    """Process one base using Viterbi path scoring with squared error between
    signal and levels plus a penalty for short dwells (number of consecutive
    stay states).

    Args:
        curr_scores (1D float array): To be populated by this function. Elements
            will be forward scores at each position
        curr_tb (1D int array): To be populated by this function. Elements will
            be the number of signal points backwards until the first point
            assigned to this base
        prev_scores (const 1D float array): Forward scores for the previous base
        curr_level (float): Expected signal level for the current base
        curr_signal (const 1D float array): Signal values for the current
            base's band
        band_start_diff (int): Difference in starting coordinates between
            current and previous base
        dwell_penalty (const float 1D array): Penalties for short dwells
    """
    cdef int band_pos, dwell_idx
    cdef float running_pos_score, pos_score

    # compute un-penalized band position scores for lookup after dwell_penalty
    # range is searched
    cdef float[::1] unpen_scores = np.empty_like(curr_scores)
    cdef int[::1] unpen_tb = np.empty_like(curr_tb)
    banded_forward_vit_step(
        unpen_scores,
        unpen_tb,
        prev_scores,
        curr_level,
        curr_signal,
        band_start_diff,
        dwell_penalty,
    )

    # loop over signal positions within this base band
    for band_pos in range(curr_scores.shape[0]):
        # if past the end of the prev band stay until the end
        if (
            band_pos + band_start_diff - prev_scores.shape[0]
            >= dwell_penalty.shape[0]
        ):
            curr_scores[band_pos] = (
                curr_scores[band_pos - 1]
                + score(curr_level, curr_signal[band_pos])
            )
            curr_tb[band_pos] = curr_tb[band_pos - 1] + 1
            continue
        # set invalid score and traceback if no valid transitions
        curr_scores[band_pos] = (
            LARGE_SCORE + prev_scores[prev_scores.shape[0] - 1]
        )
        curr_tb[band_pos] = -1
        if band_pos == 0 and band_start_diff == 0:
            continue
        running_pos_score = 0
        for dwell_idx in range(dwell_penalty.shape[0]):
            # beginning of curr or prev band reached
            if dwell_idx > band_pos or (
                band_start_diff == 0 and band_pos == dwell_idx
            ):
                break
            # update running pos score back from band_pos
            running_pos_score += score(
                curr_level, curr_signal[band_pos - dwell_idx]
            )
            # position past the end of the prev band
            if (
                band_pos - dwell_idx - 1 + band_start_diff
                >= prev_scores.shape[0]
            ):
                continue
            # compute penalized score for this band position at this dwell
            pos_score = (
                prev_scores[band_pos - dwell_idx - 1 + band_start_diff]
                + running_pos_score
                + dwell_penalty[dwell_idx]
            )
            # if this is a new min for this band pos, then update score/tb
            if pos_score < curr_scores[band_pos]:
                curr_scores[band_pos] = pos_score
                curr_tb[band_pos] = dwell_idx
        # if possible compute unpenalized score with dwell greater than
        # dwell_penalty.shape[0] using unpen_scores/unpen_tb
        if band_pos >= dwell_penalty.shape[0]:
            pos_score = (
                unpen_scores[band_pos - dwell_penalty.shape[0]]
                + running_pos_score
            )
            if pos_score < curr_scores[band_pos]:
                curr_scores[band_pos] = pos_score
                curr_tb[band_pos] = (
                    unpen_tb[band_pos - dwell_penalty.shape[0]]
                    + dwell_penalty.shape[0]
                )


cdef void banded_forward_vit_step(
    float[::1] curr_scores,
    int[::1] curr_tb,
    const float[::1] prev_scores,
    const float curr_level,
    const float[::1] curr_signal,
    int band_start_diff,
    const float[::1] sdp,
):
    """Process one base using standard Viterbi path scoring with squared error
    between signal and levels.

    Args:
        curr_scores (1D float array): To be populated by this function. Elements
            will be forward scores at each position
        curr_tb (1D int array): To be populated by this function. Elements will
            be the number of signal points backwards until the first point
            assigned to this base
        prev_scores (const 1D float array): Forward scores for the previous base
        curr_level (float): Expected signal level for the current base
        curr_signal (const 1D float array): Signal values for the current
            base's band
        band_start_diff (int): Difference in starting coordinates between
            current and previous base
        sdp (unused): Unused parameter for this core method
    """
    cdef int band_pos
    cdef float base_score, move_score, stay_score
    # compute start position in band
    if band_start_diff == 0:
        # if this is a "stay" band start, set invalid score and traceback
        curr_scores[0] = LARGE_SCORE + prev_scores[prev_scores.shape[0] - 1]
        curr_tb[0] = -1
    else:
        # else compute move score for start of base band
        base_score = score(curr_level, curr_signal[0])
        curr_scores[0] = prev_scores[band_start_diff - 1] + base_score
        curr_tb[0] = 0
        # clip prev_scores to start at same position as curr_scores
        prev_scores = prev_scores[band_start_diff:]
    # if base bands are the same
    if prev_scores.shape[0] == curr_scores.shape[0]:
        prev_scores = prev_scores[:prev_scores.shape[0] - 1]

    # compute scores where curr and prev base overlap
    for band_pos in range(1, prev_scores.shape[0] + 1):
        base_score = score(curr_level, curr_signal[band_pos])
        move_score = prev_scores[band_pos - 1] + base_score
        stay_score = curr_scores[band_pos - 1] + base_score
        if move_score < stay_score:
            curr_scores[band_pos] = move_score
            curr_tb[band_pos] = 0
        else:
            curr_scores[band_pos] = stay_score
            curr_tb[band_pos] = curr_tb[band_pos - 1] + 1

    # stay through rest of the band
    for band_pos in range(prev_scores.shape[0] + 1, curr_scores.shape[0]):
        base_score = score(curr_level, curr_signal[band_pos])
        stay_score = curr_scores[band_pos - 1] + base_score
        curr_scores[band_pos] = stay_score
        curr_tb[band_pos] = curr_tb[band_pos - 1] + 1


cdef void banded_forward_dp(
    float[::1] all_scores,
    int[::1] traceback,
    const float[::1] signal,
    const float[::1] levels,
    const int[:, ::1] seq_band,
    const int[::1] base_offsets,
    const float[::1] short_dwell_penalty,
    core_method,
):
    """Perform banded forward dynamic programming.

    Args:
        all_scores (1D int array): To be populated by this function. Ragged
            array representing each position within the band. Each element
            contains the forward score to that position within the band.
        traceback (1D int array): To be populated by this function. Ragged
            array representing each position within the band. Each element
            contains the number of signal points backwards until the first
            point assigned to this base.
        signal (1D float array): Normalized signal values.
        levels (1D float array): Expected signal levels for each base.
        seq_band (2D int array): Contains band boundaries for each base in
            signal coordinates.
        base_offsets (1D int array): Offset to the beginning of each base
        short_dwell_penalty (1D float array): Penalty values applied when
            core_method == "dwell_penalty".
        core_method (str): Core method to use for forward pass path scoring.
            "Viterbi" and "dwell_penalty" are implemented.
    """
    cdef core_func_ptr core_method_func
    if core_method == REFINE_ALGO_VIT_NAME:
        core_method_func = banded_forward_vit_step
    elif core_method == REFINE_ALGO_DWELL_PEN_NAME:
        core_method_func = banded_forward_dwell_penalty_step
    else:
        raise RemoraError(
            f"Invalid core signal mapping refine method: {core_method}"
        )

    cdef int sig_pos
    cdef int prev_bw, prev_offset, prev_band_st
    cdef int curr_bw, curr_offset, curr_band_st, curr_band_en

    # compute first base forward scores
    curr_bw = seq_band[1, 0]
    # spoof previous scores to force stays through first base
    prev_scores = np.full(curr_bw, HUGE_VALF, dtype=np.float32)
    prev_scores[0] = 0
    core_method_func(
        all_scores[:curr_bw],
        traceback[:curr_bw],
        prev_scores,
        levels[0],
        signal[:curr_bw],
        1,
        short_dwell_penalty,
    )
    prev_bw = curr_bw
    prev_band_st = prev_offset = 0

    # compute forward scores for all bases
    for base_idx in range(1, levels.shape[0]):
        curr_band_st = seq_band[0, base_idx]
        curr_band_en = seq_band[1, base_idx]
        curr_bw = curr_band_en - curr_band_st
        curr_offset = base_offsets[base_idx]
        # compute all scores and traceback for this base
        core_method_func(
            all_scores[curr_offset:curr_offset + curr_bw],
            traceback[curr_offset:curr_offset + curr_bw],
            all_scores[prev_offset:prev_offset + prev_bw],
            levels[base_idx],
            signal[curr_band_st:curr_band_en],
            curr_band_st - prev_band_st,
            short_dwell_penalty,
        )
        prev_band_st = curr_band_st
        prev_bw = curr_bw
        prev_offset = curr_offset


def seq_banded_dp(
    const float[::1] signal,
    const float[::1] levels,
    const int[:, ::1] seq_band,
    const float[::1] short_dwell_penalty,
    core_method=REFINE_ALGO_VIT_NAME,
):
    """Decode the path assignment between the input signal and levels.

    Note that paths are restricted to the input seq band.

    Args:
        signal (np.ndarray): Float32 array containing normalized signal values.
        levels (np.ndarray): Float32 array containing estimated levels.
        seq_band (np.ndarray): int32 np.ndarray with shape
            (2, seq_len = levels.shape[0]). The first row contains the lower
            band boundaries in signal coordinates and the second row contains
            the upper boundaries in signal coordinates. The first lower bound
            should be 0 and last upper bound should be signal.shape[0].
        short_dwell_penalty (np.ndarray): Float32 array with penalty values for
            short dwells. Length of the array defines the length of dwells that
            are penalized.
        core_method (str): Perform decoding with one of the following options:
            - "Viterbi": Standard Viterbi forward pass
            - "dwell_penalty": Viterbi forward pass using short_dwell_penalty

    Returns:
        Tuple containing:
            1)  Float32 1D numpy array representing each position within
                the band. Signal positions within each base are stored
                contiguously. Each element contains the forward score to that
                position within the band.
            2) Int32 np.ndarray of length = level.shape[0] + 1. The first value
               will be zero, and the remaining values are the start locations
               of each level within the signal array.
    """
    # Prepare for banded forwards pass followed by traceback
    base_offsets = np.empty(seq_band.shape[1] + 1, dtype=np.int32)
    base_offsets[0] = 0
    base_offsets[1:] = np.cumsum(np.diff(seq_band, axis=0)[0])
    band_len = base_offsets[base_offsets.shape[0] - 1]

    # all_scores array holds current score at every sequence by signal position
    # within the band
    all_scores = np.empty(band_len, dtype=np.float32)
    # traceback contains one value for each position in the band and contains
    # the number of stays from this position until the next move back to the
    # previous base/level.
    traceback = np.empty(band_len, dtype=np.int32)
    # Execute forward pass filling all_scores and traceback
    banded_forward_dp(
        all_scores,
        traceback,
        signal,
        levels,
        seq_band,
        base_offsets,
        short_dwell_penalty,
        core_method,
    )
    # path is primary return value as described in the return type
    path = np.empty(levels.shape[0] + 1, dtype=np.int32)
    # perform traceback and full path
    banded_traceback(path, seq_band, base_offsets, traceback)
    return all_scores, path, traceback, base_offsets


def forward_dp(
    const float[::1] signal,
    const float[::1] levels,
    const int[:, ::1] seq_band,
    const float[::1] short_dwell_penalty,
    core_method=REFINE_ALGO_VIT_NAME,
):
    base_offsets = np.empty(seq_band.shape[1] + 1, dtype=np.int32)
    base_offsets[0] = 0
    base_offsets[1:] = np.cumsum(np.diff(seq_band, axis=0)[0])
    band_len = base_offsets[base_offsets.shape[0] - 1]
    all_scores = np.empty(band_len, dtype=np.float32)
    traceback = np.empty(band_len, dtype=np.int32)
    banded_forward_dp(
        all_scores,
        traceback,
        signal,
        levels,
        seq_band,
        base_offsets,
        short_dwell_penalty,
        core_method,
    )
    return all_scores, traceback, base_offsets



import os
import argparse
import warnings

from remora import __version__

# from remora.common import logging
from remora.parsers import (
    register_dataset,
    register_model,
    register_infer,
    register_validate,
    register_analyze,
    SubcommandHelpFormatter,
)

# filter warnings about GPU on environments without a GPU
warnings.filterwarnings(action="ignore", category=UserWarning, module="torch")

# LOGGER = logging.get_logger()

_DO_PROFILE = False
# None if environment var not set
_PROF_FN = os.getenv("REMORA_PROFILE_FILE")
if _PROF_FN:
    _DO_PROFILE = True
    # LOGGER.warning(f"Profiling remora. Saving profile data to {_PROF_FN}")


def run():
    """The main routine."""
    # prepare first level `remora -h` help, including description.
    desc = (
        "********** Remora *********\n\nModified base model training and "
        "application.\n\n"
    )
    parser = argparse.ArgumentParser(
        prog="remora",
        description=desc,
        formatter_class=SubcommandHelpFormatter,
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="Remora version: {}".format(__version__),
        help="Show Remora version and exit.",
    )
    parser.set_defaults(func=lambda x: parser.print_help())

    subparsers = parser.add_subparsers(title="sub-commands")
    register_dataset(subparsers)
    register_model(subparsers)
    register_infer(subparsers)
    register_validate(subparsers)
    register_analyze(subparsers)

    args = parser.parse_args()
    cmd_func = args.func
    if _DO_PROFILE:
        _func_wrapper = cmd_func

        def cmd_func(args):
            import cProfile

            prof = cProfile.Profile()
            retval = prof.runcall(_func_wrapper, args)
            prof.dump_stats(_PROF_FN)
            return retval

    cmd_func(args)


if __name__ == "__main__":
    run()



import os
import sys
import json
import toml
import datetime
import importlib
import pkg_resources
from pathlib import Path
from os.path import isfile

import torch
import numpy as np
import pandas as pd
from torch import nn
from torch.nn.utils.fusion import fuse_conv_bn_eval

from remora.download import ModelDownload
from remora import log, RemoraError, constants
from remora.refine_signal_map import SigMapRefiner

LOGGER = log.get_logger()

#############
# Exporting #
#############


def export_model_torchscript(ckpt, model, save_filename):
    model.eval()
    m = torch.jit.script(model)
    meta = {}
    meta["creation_date"] = datetime.datetime.now().strftime(
        "%m/%d/%Y, %H:%M:%S"
    )
    meta["kmer_context_bases"] = ckpt["kmer_context_bases"]
    meta["chunk_context"] = ckpt["chunk_context"]

    # add simple metadata
    for ckpt_key in (
        "base_pred",
        "mod_bases",
        "reverse_signal",
        "refine_kmer_center_idx",
        "refine_do_rough_rescale",
        "refine_scale_iters",
        "refine_algo",
        "refine_half_bandwidth",
        "base_start_justify",
        "offset",
        "model_params",
    ):
        meta[ckpt_key] = ckpt[ckpt_key]

    if ckpt["mod_bases"] is not None:
        for mod_idx in range(len(ckpt["mod_bases"])):
            meta[f"mod_long_names_{mod_idx}"] = str(
                ckpt["mod_long_names"][mod_idx]
            )
    meta["num_motifs"] = str(ckpt["num_motifs"])
    for idx, (motif, motif_offset) in enumerate(ckpt["motifs"]):
        m_key = f"motif_{idx}"
        mo_key = f"motif_offset_{idx}"
        meta[m_key] = str(motif)
        meta[mo_key] = str(motif_offset)

    meta["num_motifs"] = str(len(ckpt["motifs"]))

    # store refine arrays as bytes
    meta["refine_kmer_levels"] = (
        ckpt["refine_kmer_levels"].astype(np.float32).tobytes().decode("cp437")
    )
    meta["refine_sd_arr"] = (
        ckpt["refine_sd_arr"].astype(np.float32).tobytes().decode("cp437")
    )
    meta["doc_string"] = "Nanopore Remora model"
    try:
        meta["model_version"] = ckpt["model_version"]
    except KeyError:
        LOGGER.warning("Model version not found in checkpoint. Setting to 0.")
        meta["model_version"] = 0
    json_object = json.dumps(meta, indent=4)
    extra_files = {"meta.txt": json_object}
    torch.jit.save(m, save_filename, _extra_files=extra_files)


def export_model_dorado(ckpt, model, save_dir):
    save_dir = os.path.expanduser(save_dir)
    if os.path.exists(save_dir):
        LOGGER.info(
            f'Directory "{save_dir}" already exists. Exported files will '
            "be overwritten."
        )
    os.makedirs(save_dir, exist_ok=True)

    def save_tensor(fn, x):
        m = torch.nn.Module()
        par = nn.Parameter(x, requires_grad=False)
        m.register_parameter("0", par)
        tensors = torch.jit.script(m)
        tensors.save(f"{save_dir}/{fn}.tensor")
        LOGGER.info(f"{save_dir}/{fn}.tensor")

    model.eval()
    layer_names = set()

    conv_to_bn = {
        "sig_conv1": "sig_bn1",
        "sig_conv2": "sig_bn2",
        "sig_conv3": "sig_bn3",
        "seq_conv1": "seq_bn1",
        "seq_conv2": "seq_bn2",
        "seq_conv3": "seq_bn3",
        "merge_conv1": "merge_bn",
        "merge_conv2": "merge_bn2",
        "merge_conv3": "merge_bn3",
        "merge_conv4": "merge_bn4",
    }

    for name, module in model.named_modules():
        if name == "" or "bn" in name or "drop" in name:
            continue
        if name in conv_to_bn.keys():
            fused = fuse_conv_bn_eval(module, getattr(model, conv_to_bn[name]))
            module.weight = fused.weight
            module.bias = fused.bias
        for k, v in module.state_dict().items():
            save_tensor(f"{name}.{k}", v)
        layer_names.add(name)

    metadata = {}
    general = {}
    modbases = {}

    general["creation_date"] = datetime.datetime.now().strftime(
        "%m/%d/%Y, %H:%M:%S"
    )

    lstm_model_layers = {
        "sig_conv1",
        "sig_conv2",
        "sig_conv3",
        "seq_conv1",
        "seq_conv2",
        "merge_conv1",
        "lstm1",
        "lstm2",
        "fc",
    }

    conv_model_layers = {
        "sig_conv1",
        "sig_conv2",
        "sig_conv3",
        "seq_conv1",
        "seq_conv2",
        "seq_conv3",
        "merge_conv1",
        "merge_conv2",
        "merge_conv3",
        "merge_conv4",
        "fc",
    }

    if layer_names == conv_model_layers:
        general["model"] = "conv_only"
    elif layer_names == lstm_model_layers:
        general["model"] = "conv_lstm"
    else:
        LOGGER.warning("Unknown layer setup in export")
        general["model"] = "unknown"

    # add refinement metadata
    refinement = {}
    refinement["refine_do_rough_rescale"] = ckpt["refine_do_rough_rescale"]
    if refinement["refine_do_rough_rescale"]:
        refinement["refine_kmer_center_idx"] = ckpt["refine_kmer_center_idx"]
        try:
            refine_kmer_levels = torch.Tensor(
                ckpt["refine_kmer_levels"].astype(np.float32)
            )
        except AttributeError:
            refine_kmer_levels = torch.from_numpy(
                np.frombuffer(
                    ckpt["refine_kmer_levels"].encode("cp437"),
                    dtype=np.float32,
                ).copy()
            )
        save_tensor("refine_kmer_levels", refine_kmer_levels)

    # add simple metadata
    for ckpt_key in (
        "mod_bases",
        "offset",
        "reverse_signal",
    ):
        modbases[ckpt_key] = ckpt[ckpt_key]

    if ckpt["mod_bases"] is not None:
        for mod_idx in range(len(ckpt["mod_bases"])):
            modbases[f"mod_long_names_{mod_idx}"] = str(
                ckpt["mod_long_names"][mod_idx]
            )
    for key in ("chunk_context", "kmer_context_bases"):
        for idx in range(2):
            modbases[f"{key}_{idx}"] = ckpt[key][idx]

    if len(ckpt["motifs"]) > 1:
        raise RemoraError("Dorado only supports models with a single motif")

    for motif, motif_offset in ckpt["motifs"]:
        modbases["motif"] = motif
        modbases["motif_offset"] = motif_offset

    metadata["general"] = general
    metadata["model_params"] = ckpt["model_params"]
    metadata["modbases"] = modbases
    metadata["refinement"] = refinement

    toml.dump(metadata, open(os.path.join(save_dir, "config.toml"), "w"))


###########
# Loading #
###########


def _load_python_model(model_file, **model_kwargs):
    loader = importlib.machinery.SourceFileLoader("netmodule", model_file)
    netmodule = importlib.util.module_from_spec(
        importlib.util.spec_from_loader(loader.name, loader)
    )
    loader.exec_module(netmodule)
    network = netmodule.network(**model_kwargs)
    return network


def continue_from_checkpoint(ckp_path, model_path=None):
    """Load a checkpoint in order to continue training."""
    if not isfile(ckp_path):
        raise RemoraError(f"Checkpoint path is not a file ({ckp_path})")
    ckpt = torch.load(ckp_path, map_location="cpu")
    if ckpt["state_dict"] is None:
        raise RemoraError("Model state not saved in checkpoint.")

    model_path = ckpt["model_path"] if model_path is None else model_path
    model = _load_python_model(model_path, **ckpt["model_params"])
    model.load_state_dict(ckpt["state_dict"])
    return ckpt, model


def add_derived_metadata(model_metadata):
    if "reverse_signal" not in model_metadata:
        LOGGER.warning(
            "reverse signal attribute not found in model. Assuming False"
        )
        model_metadata["reverse_signal"] = False
    if model_metadata["mod_bases"] == "None":
        model_metadata["mod_bases"] = None
        model_metadata["mod_long_names"] = None
    else:
        model_metadata["mod_long_names"] = []
        for mod_idx in range(len(model_metadata["mod_bases"])):
            model_metadata["mod_long_names"].append(
                model_metadata[f"mod_long_names_{mod_idx}"]
            )
    if "kmer_context_bases" not in model_metadata:
        model_metadata["kmer_context_bases"] = (
            int(model_metadata["kmer_context_bases_0"]),
            int(model_metadata["kmer_context_bases_1"]),
        )
    model_metadata["kmer_len"] = sum(model_metadata["kmer_context_bases"]) + 1
    if "chunk_context" not in model_metadata:
        model_metadata["chunk_context"] = (
            int(model_metadata["chunk_context_0"]),
            int(model_metadata["chunk_context_1"]),
        )
    model_metadata["chunk_len"] = sum(model_metadata["chunk_context"])

    if "num_motifs" not in model_metadata:
        model_metadata["motifs"] = [
            (model_metadata["motif"], int(model_metadata["motif_offset"]))
        ]
        model_metadata["motif_offset"] = int(model_metadata["motif_offset"])
    else:
        num_motifs = int(model_metadata["num_motifs"])

        motifs = []
        motif_offsets = []
        for mot in range(num_motifs):
            motifs.append(model_metadata[f"motif_{mot}"])
            motif_offsets.append(model_metadata[f"motif_offset_{mot}"])

        model_metadata["motifs"] = [
            (mot, int(mot_off)) for mot, mot_off in zip(motifs, motif_offsets)
        ]

    model_metadata["can_base"] = model_metadata["motifs"][0][0][
        model_metadata["motifs"][0][1]
    ]

    # allowed settings for this attribute are a single motif or all-contexts
    # note that inference core methods use motifs attribute
    if len(model_metadata["motifs"]) == 1:
        model_metadata["motif"] = model_metadata["motifs"][0]
    else:
        model_metadata["motif"] = (model_metadata["can_base"], 0)

    mod_str = "; ".join(
        f"{mod_b}={mln}"
        for mod_b, mln in zip(
            model_metadata["mod_bases"], model_metadata["mod_long_names"]
        )
    )
    model_metadata["alphabet_str"] = (
        "loaded modified base model to call (alt to "
        f"{model_metadata['can_base']}): {mod_str}"
    )

    if "refine_kmer_levels" in model_metadata:
        # load sig_map_refiner
        levels_array = np.frombuffer(
            model_metadata["refine_kmer_levels"].encode("cp437"),
            dtype=np.float32,
        )
        model_metadata["refine_kmer_levels"] = levels_array
        refine_sd_arr = np.frombuffer(
            model_metadata["refine_sd_arr"].encode("cp437"), dtype=np.float32
        )
        model_metadata["refine_sd_arr"] = refine_sd_arr
        model_metadata["sig_map_refiner"] = SigMapRefiner(
            _levels_array=levels_array,
            center_idx=int(model_metadata["refine_kmer_center_idx"]),
            do_rough_rescale=bool(
                int(model_metadata["refine_do_rough_rescale"])
            ),
            scale_iters=int(model_metadata["refine_scale_iters"]),
            algo=model_metadata["refine_algo"],
            half_bandwidth=int(model_metadata["refine_half_bandwidth"]),
            sd_arr=refine_sd_arr,
        )
    else:
        # handle original models without sig_map_refiner
        model_metadata["sig_map_refiner"] = SigMapRefiner()
        model_metadata["base_start_justify"] = False
        model_metadata["offset"] = 0


def repr_model_metadata(metadata):
    # skip attributes included in parsed values
    return "\n".join(
        f"  {k: >20} : {v}"
        for k, v in metadata.items()
        if not any(
            k.startswith(val)
            for val in (
                "mod_long_names_",
                "kmer_context_bases_",
                "chunk_context_",
                "motif_",
                "refine_kmer_levels",
                "refine_sd_arr",
                "refine_kmer_center_idx",
                "refine_do_rough_rescale",
                "refine_scale_iters",
                "refine_algo",
                "refine_half_bandwidth",
            )
        )
    )


def load_torchscript_model(
    model_filename, device=None, quiet=False, eval_only=False
):
    """Load torchscript model. If device is specified load onto specified
    device.

    Args:
        model_filename (str): Model path
        device (torch.device): Torch device (or None)
        quiet (bool): Print model info to debug
        eval_only (bool): Load model in eval mode and requires_grad=False. Note
            that torch.set_grad_enabled(False) should be set as well for
            optimal inference performance.

    Returns:
        2-tuple containing:
          1. Compiled model object for calling mods
          2. Model metadata dictionary with information concerning data prep
    """

    # values will be replaced with data
    extra_files = {"meta.txt": ""}
    if device is None:
        model = torch.jit.load(
            model_filename, _extra_files=extra_files, map_location="cpu"
        )
    else:
        model = torch.jit.load(
            model_filename,
            _extra_files=extra_files,
            map_location=device,
        )
    model_metadata = json.loads(extra_files["meta.txt"])
    add_derived_metadata(model_metadata)
    if not quiet:
        md_str = repr_model_metadata(model_metadata)
        LOGGER.debug(f"Loaded Remora model attrs\n{md_str}\n")
    if eval_only:
        model.eval()
        for param in model.parameters():
            param.requires_grad = False
    return model, model_metadata


def load_model(
    model_filename=None,
    *,
    pore=None,
    basecall_model_type=None,
    basecall_model_version=None,
    modified_bases=None,
    remora_model_type=None,
    remora_model_version=None,
    device=None,
    quiet=True,
    eval_only=False,
):
    if model_filename is not None:
        if not isfile(model_filename):
            raise RemoraError(
                f"Remora model file ({model_filename}) not found."
            )
        try:
            LOGGER.debug("Using torchscript model")
            return load_torchscript_model(
                model_filename, device, quiet=quiet, eval_only=eval_only
            )
        except (AttributeError, RuntimeError):
            raise RemoraError("Failed loading torchscript model.")

    if pore is None:
        raise RemoraError("Must specify a pore.")
    try:
        pore = pore.lower()
        submodels = constants.MODEL_DICT[pore]
    except (AttributeError, KeyError):
        pores = ", ".join(constants.MODEL_DICT.keys())
        raise RemoraError(
            f"No trained Remora models for {pore}. Options: {pores}"
        )

    if modified_bases is None:
        raise RemoraError("Must specify a modified base.")
    try:
        modified_bases = "_".join(sorted(x.lower() for x in modified_bases))
        submodels = submodels[modified_bases]
    except (AttributeError, KeyError):
        raise RemoraError(
            f"Remora model for modified bases {modified_bases} not found "
            f"for {pore}"
        )

    if remora_model_type is None:
        remora_model_type = next(iter(submodels.items()))[0]
        LOGGER.info(
            "Modified bases model type not supplied. Using default "
            f"{remora_model_type}."
        )
    try:
        submodels = submodels[remora_model_type]
    except (AttributeError, KeyError):
        LOGGER.warning(
            f"Remora model type {remora_model_type} not found "
            f"for {pore} {modified_bases}. Using default Remora model."
        )
        remora_model_type = next(iter(submodels.items()))[0]
        submodels = submodels[remora_model_type]

    if basecall_model_type is None:
        basecall_model_type = next(iter(submodels.items()))[0]
        LOGGER.info(
            "Basecaller model type not supplied. Using default "
            f"{basecall_model_type}."
        )
    try:
        basecall_model_type = basecall_model_type.lower()
        submodels = submodels[basecall_model_type]
    except (AttributeError, KeyError):
        LOGGER.warning(
            f"No trained Remora models for {basecall_model_type} "
            f"(with {pore}). Using default Remora model."
        )
        basecall_model_type = next(iter(submodels.items()))[0]
        submodels = submodels[basecall_model_type]

    if basecall_model_version is None:
        basecall_model_version = next(iter(submodels.items()))[0]
        LOGGER.info(
            "Basecall model version not supplied. Using default Remora model "
            f"for {pore}_{basecall_model_type}."
        )
    try:
        submodels = submodels[basecall_model_version]
    except KeyError:
        LOGGER.warning(
            "Remora model for basecall model version "
            f"({basecall_model_version}) not found. Using default Remora "
            f"model for {pore}_{basecall_model_type}."
        )
        basecall_model_version = next(iter(submodels.items()))[0]
        submodels = submodels[basecall_model_version]

    if remora_model_version is None:
        LOGGER.info("Remora model version not specified. Using latest.")
        remora_model_version = next(iter(submodels.items()))[0]
    if remora_model_version not in submodels:
        LOGGER.warning(
            f"Remora model version {remora_model_version} not found. "
            "Using latest."
        )
        remora_model_version = next(iter(submodels.items()))[0]

    try:
        url = submodels[remora_model_version]
    except (AttributeError, KeyError):
        raise RemoraError(
            "Remora model url not found "
            f"for {pore}_{basecall_model_type}@{basecall_model_version} "
            f"{modified_bases}_{remora_model_type}_v{remora_model_version}."
        )

    remora_model_version = f"v{remora_model_version}"
    path = pkg_resources.resource_filename(
        "remora",
        constants.MODEL_DATA_DIR_NAME,
    )
    path = Path(path)
    path.mkdir(parents=True, exist_ok=True)
    model_name = "_".join(
        [
            pore,
            basecall_model_type,
            basecall_model_version,
            modified_bases,
            remora_model_type,
            remora_model_version,
        ]
    )
    model_name = f"{model_name}.pt"
    full_path = os.path.join(path, model_name)
    if not os.path.exists(full_path):
        LOGGER.info(
            f"No pre-trained Remora model found for "
            f"this configuration {model_name} at {path}.\n"
            f"Attempting to download {model_name}"
        )
        md = ModelDownload(path)
        md.download(url)
    try:
        return load_torchscript_model(full_path, device, eval_only=eval_only)
    except (AttributeError, RuntimeError):
        raise RemoraError("Failed loading torchscript model.")


def get_pretrained_models(
    pore=None,
    basecall_model_type=None,
    basecall_model_version=None,
    modified_bases=None,
    remora_model_type=None,
    remora_model_version=None,
):
    def filter_dataframe(models, args):
        running_sequence = []
        for x, y in zip(args, models.columns):
            if x is None:
                continue
            else:
                if y in ["Pore", "Basecall_Model_Type"]:
                    x = x.lower()
                elif y in ["Modified_Bases"]:
                    x = "_".join(sorted(z.lower() for z in x))
                elif y in ["Remora_Model_Type"]:
                    x = x.upper()
                running_sequence.append(x)
                models = models[models[y] == x]
                if len(models) == 0:
                    LOGGER.info(
                        f" {y} {','.join(running_sequence)} not found in "
                        "library of pre-trained models."
                    )
                    sys.exit(1)

        return models

    header = [
        "Pore",
        "Basecall_Model_Type",
        "Basecall_Model_Version",
        "Modified_Bases",
        "Remora_Model_Type",
        "Remora_Model_Version",
        "Remora_Model_URL",
    ]

    models = []
    for pore_type, mod_bases in constants.MODEL_DICT.items():
        for mod_base, remora_types in mod_bases.items():
            for remora_type, bc_types in remora_types.items():
                for bc_type, bc_vers in bc_types.items():
                    for bc_ver, remora_vers in bc_vers.items():
                        for remora_ver, remora_url in remora_vers.items():
                            models.append(
                                (
                                    pore_type,
                                    bc_type,
                                    bc_ver,
                                    mod_base,
                                    remora_type,
                                    remora_ver,
                                    remora_url,
                                )
                            )
    models = pd.DataFrame(models)
    models.columns = header

    args = [
        pore,
        basecall_model_type,
        basecall_model_version,
        modified_bases,
        remora_model_type,
        remora_model_version,
    ]

    models = filter_dataframe(models, args)

    return models, header



import sys
import logging


class CustomFormatter(logging.Formatter):
    """Custom formatter for python logging module.

    Creates separate logging format for error, warning, info and debug logging
    levels.
    """

    err_str = (
        "[%(asctime)s:%(processName)s:%(threadName)s:%(module)s.py:%(lineno)d]"
    )
    err_fmt = "*" * 100 + f"\n\tERROR {err_str}:\n%(msg)s\n" + "*" * 100
    warn_fmt = "*" * 20 + f" WARNING {err_str}: %(msg)s " + "*" * 20
    info_fmt = "[%(asctime)s] %(message)s"
    dbg_fmt = f"DEBUG {err_str} %(msg)s"

    def __init__(self, fmt="[%(asctime)s] %(message)s"):
        super().__init__(fmt=fmt, datefmt="%H:%M:%S", style="%")

    def format(self, record):
        format_orig = self._fmt

        # Replace the original format with one customized by logging level
        if record.levelno == logging.DEBUG:
            self._style._fmt = self.dbg_fmt
        elif record.levelno == logging.INFO:
            self._style._fmt = self.info_fmt
        elif record.levelno == logging.WARNING:
            self._style._fmt = self.warn_fmt
        elif record.levelno == logging.ERROR:
            self._style._fmt = self.err_fmt
        result = logging.Formatter.format(self, record)

        self._fmt = format_orig

        return result


CONSOLE = logging.StreamHandler()
CONSOLE.setLevel(logging.INFO)
CONSOLE.setFormatter(CustomFormatter())
ROOT_LOGGER = logging.getLogger("Remora")
ROOT_LOGGER.setLevel(logging.DEBUG)
ROOT_LOGGER.addHandler(CONSOLE)


def init_logger(log_fn=None, quiet=False):
    """Prepare logging output.

    Args:
        log_fn (str): Path to logging output file. All logging messages,
            including debug level, will be output to this file.
        quiet (bool): Set console logging level to warning. Default info.
    """
    log_fp = None
    if log_fn is not None:
        log_fp = logging.FileHandler(log_fn, mode="a")
        log_fp.setLevel(logging.DEBUG)
        log_fp.setFormatter(CustomFormatter())
    if log_fp is not None:
        ROOT_LOGGER.addHandler(log_fp)
    if quiet:
        CONSOLE.setLevel(logging.WARNING)
    logging.getLogger("Remora").debug(f'Command: """{" ".join(sys.argv)}"""')


def get_logger(module_name="Remora"):
    return logging.getLogger(module_name)


if __name__ == "__main__":
    sys.stderr.write("This is a module. See commands with `remora -h`")
    sys.exit(1)



import os

import requests
from tqdm import tqdm

from remora import log, RemoraError

LOGGER = log.get_logger()


class ModelDownload:
    __url__ = "https://cdn.oxfordnanoportal.com/software/analysis/remora/"

    def __init__(self, path, force=False):
        self.path = path
        self.force = force

    def location(self, filename):
        return os.path.join(self.path, filename)

    def exists(self, filename):
        return os.path.exists(self.location(filename))

    def download(self, url_frag):
        filename = f"{url_frag}.pt"
        url = f"{self.__url__}{filename}"
        req = requests.get(url, stream=True)
        total = int(req.headers.get("content-length", 0))

        if self.exists(filename) and not self.force:
            LOGGER.info(
                f"Model already exists, skipping download of {filename}"
            )
            return

        with tqdm(
            total=total,
            unit="iB",
            ncols=100,
            unit_scale=True,
            disable=os.environ.get("LOG_SAFE", False),
        ) as t:
            try:
                with open(self.location(filename), "wb") as f:
                    for data in req.iter_content(1024):
                        f.write(data)
                        t.update(len(data))
            except OSError as e:
                raise RemoraError(
                    "File-system was unable to write download model. "
                    f"URL is {url} and intended download path was "
                    f"{self.location(filename)}.\nFull error message: {e}"
                )
        LOGGER.info(f"Model {filename} downloaded to {self.location(filename)}")



from collections import deque
from typing import List, Tuple
from dataclasses import dataclass

import parasail
from numpy import typing as npt

from remora import data_chunks as DC

CigarTuples = List[Tuple[int, int]]


@dataclass
class PairwiseAlignment:
    ref_start: int
    ref_end: int
    query_start: int
    query_end: int
    cigar: CigarTuples


def trim_parasail_alignment(alignment_result):
    ref_start = 0
    ref_end = alignment_result.len_ref
    query_start = 0
    query_end = alignment_result.len_query
    fixed_start = False
    fixed_end = False

    cigar_string = alignment_result.cigar.decode.decode()
    cigar_tuples = deque(DC.cigartuples_from_string(cigar_string))
    while not (fixed_start and fixed_end):
        fist_op, first_length = cigar_tuples[0]
        if fist_op in (1, 4):  # insert, soft-clip, increment query start
            query_start += first_length
            cigar_tuples.popleft()
        elif fist_op == 2:  # delete, increment reference start
            ref_start += first_length
            cigar_tuples.popleft()
        else:
            fixed_start = True

        last_op, last_length = cigar_tuples[-1]
        if last_op in (1, 4):  # decrement the query end
            query_end -= last_length
            cigar_tuples.pop()
        elif last_op == 2:  # decrement the ref_end
            ref_end -= last_length
            cigar_tuples.pop()
        else:
            fixed_end = True

    return PairwiseAlignment(
        ref_start=ref_start,
        ref_end=ref_end,
        query_start=query_start,
        query_end=query_end,
        cigar=list(cigar_tuples),
    )


def parasail_align(*, query, ref) -> PairwiseAlignment:
    """
    Semi-global alignment allowing for gaps at the start and end of the query
    sequence.

    :param query: str
    :param ref: str
    :return: PairwiseAlignment
    :raises RuntimeError when no matching/mismatching operations are found
    """
    alignment_result = parasail.sg_qx_trace_scan_32(
        query, ref, 10, 2, parasail.dnafull
    )
    assert alignment_result.len_ref == len(
        ref
    ), "alignment reference length is discordant with reference"
    assert alignment_result.len_query == len(
        query
    ), "alignment query length is discordant with query"
    assert alignment_result.end_ref == len(ref) - 1, "end_ref is not the end"

    try:
        return trim_parasail_alignment(alignment_result)
    except IndexError as e:
        raise RuntimeError(
            "failed to find match operations in pairwise alignment"
        ) from e


@dataclass
class SimplexDuplexMapping:
    duplex_to_simplex_mapping: npt.NDArray[int]
    trimmed_duplex_seq: str
    duplex_offset: int


def map_simplex_to_duplex(*, simplex_seq: str, duplex_seq: str):
    pairwise_alignment = parasail_align(query=simplex_seq, ref=duplex_seq)

    trimmed_duplex = duplex_seq[
        pairwise_alignment.ref_start : pairwise_alignment.ref_end
    ]
    duplex_to_simplex_mapping = (
        DC.make_sequence_coordinate_mapping(
            pairwise_alignment.cigar,
        ).astype(int)
        + pairwise_alignment.query_start
    )

    return SimplexDuplexMapping(
        duplex_to_simplex_mapping=duplex_to_simplex_mapping,
        trimmed_duplex_seq=trimmed_duplex,
        duplex_offset=pairwise_alignment.ref_start,
    )



import os
import atexit
from shutil import copyfile

import torch
import numpy as np
from tqdm import tqdm
from thop import profile

from remora.data_chunks import RemoraDataset
from remora import (
    constants,
    util,
    log,
    RemoraError,
    encoded_kmers,
    model_util,
    validate,
)

LOGGER = log.get_logger()
BREACH_THRESHOLD = 0.8
REGRESSION_THRESHOLD = 0.7


def load_optimizer(optimizer, model, lr, weight_decay, momentum=0.9):
    if optimizer == constants.SGD_OPT:
        return torch.optim.SGD(
            model.parameters(),
            lr=lr,
            weight_decay=weight_decay,
            momentum=momentum,
            nesterov=True,
        )
    elif optimizer == constants.ADAM_OPT:
        return torch.optim.Adam(
            model.parameters(),
            lr=lr,
            weight_decay=weight_decay,
        )
    elif optimizer == constants.ADAMW_OPT:
        return torch.optim.AdamW(
            model.parameters(),
            lr=lr,
            weight_decay=weight_decay,
        )
    raise RemoraError(f"Invalid optimizer specified ({optimizer})")


def select_scheduler(scheduler, opt, lr_sched_kwargs):
    lr_sched_dict = {}
    if lr_sched_kwargs is None and scheduler is None:
        scheduler = constants.DEFAULT_SCHEDULER
        lr_sched_dict = constants.DEFAULT_SCH_VALUES
    else:
        for lr_key, lr_val, lr_type in lr_sched_kwargs:
            lr_sched_dict[lr_key] = constants.TYPE_CONVERTERS[lr_type](lr_val)
    return getattr(torch.optim.lr_scheduler, scheduler)(opt, **lr_sched_dict)


def save_model(
    model,
    ckpt_save_data,
    out_path,
    epoch,
    opt,
    model_name=constants.BEST_MODEL_FILENAME,
    as_torchscript=True,
    model_name_torchscript=constants.BEST_TORCHSCRIPT_MODEL_FILENAME,
):
    ckpt_save_data["epoch"] = epoch + 1
    state_dict = model.state_dict()
    if "total_ops" in state_dict.keys():
        state_dict.pop("total_ops", None)
    if "total_params" in state_dict.keys():
        state_dict.pop("total_params", None)
    ckpt_save_data["state_dict"] = state_dict
    ckpt_save_data["opt"] = opt.state_dict()
    torch.save(
        ckpt_save_data,
        os.path.join(out_path, model_name),
    )
    if as_torchscript:
        model_util.export_model_torchscript(
            ckpt_save_data,
            model,
            os.path.join(out_path, model_name_torchscript),
        )


def train_model(
    seed,
    device,
    out_path,
    remora_dataset_path,
    chunk_context,
    kmer_context_bases,
    val_prop,
    batch_size,
    model_path,
    size,
    optimizer,
    lr,
    scheduler_name,
    weight_decay,
    epochs,
    save_freq,
    early_stopping,
    filt_frac,
    ext_val,
    ext_val_names,
    lr_sched_kwargs,
    balance,
    batch_label_weights,
    high_conf_incorrect_thr_frac,
    finetune_path,
    freeze_num_layers,
):
    seed = (
        np.random.randint(0, np.iinfo(np.uint32).max, dtype=np.uint32)
        if seed is None
        else seed
    )
    LOGGER.info(f"Seed selected is {seed}")

    np.random.seed(seed)
    torch.manual_seed(seed)
    if device is not None and device.type == "cuda":
        torch.cuda.manual_seed_all(seed)
        torch.cuda.set_device(device)

    LOGGER.info("Loading dataset from Remora file")
    dataset = RemoraDataset.load_from_file(
        remora_dataset_path,
        batch_size=batch_size,
        drop_read_attrs=True,
    )
    LOGGER.info(f"Dataset loaded with labels: {dataset.get_label_counts()}")
    if balance:
        dataset = dataset.balance_classes()
        LOGGER.info(f"Dataset balanced: {dataset.get_label_counts()}")
    dataset.trim_kmer_context_bases(kmer_context_bases)
    dataset.trim_chunk_context(chunk_context)
    # load attributes from file
    LOGGER.info(f"Dataset summary:\n{dataset.summary}")

    val_fp = open(out_path / "validation.log", mode="w", buffering=1)
    atexit.register(val_fp.close)
    val_fp = validate.ValidationLogger(val_fp)
    batch_fp = open(out_path / "batch.log", "w", buffering=1)
    atexit.register(batch_fp.close)
    if high_conf_incorrect_thr_frac is not None:
        batch_fp.write("Iteration\tLoss\tNumberFiltered\n")
    else:
        batch_fp.write("Iteration\tLoss\n")

    LOGGER.info("Loading model")
    copy_model_path = util.resolve_path(os.path.join(out_path, "model.py"))
    copyfile(model_path, copy_model_path)
    num_out = 4 if dataset.base_pred else len(dataset.mod_bases) + 1
    model_params = {
        "size": size,
        "kmer_len": sum(dataset.kmer_context_bases) + 1,
        "num_out": num_out,
    }
    model = model_util._load_python_model(copy_model_path, **model_params)
    LOGGER.info(f"Model structure:\n{model}")

    if finetune_path:
        ckpt, model = model_util.continue_from_checkpoint(
            finetune_path, copy_model_path
        )
        if freeze_num_layers:
            for freeze_iter, (p_name, param) in enumerate(
                model.named_parameters()
            ):
                LOGGER.debug(f"Freezing layer for training: {p_name}")
                param.requires_grad = False
                if freeze_iter >= freeze_num_layers:
                    break

        if ckpt["model_params"]["num_out"] != num_out:
            in_feat = model.fc.in_features
            model.fc = torch.nn.Linear(in_feat, num_out)
        if ckpt["model_params"]["size"] != size:
            LOGGER.warning(
                "Size mismatch between pretrained model and selected size. "
                "Using pretrained model size."
            )
            model_params["size"] = ckpt["model_params"]["size"]
        if dataset.chunk_context != ckpt["chunk_context"]:
            raise RemoraError(
                "The chunk context of the pre-trained model and the dataset "
                "do not match."
            )
        if dataset.kmer_context_bases != ckpt["kmer_context_bases"]:
            raise RemoraError(
                "The kmer context bases of the pre-trained model and "
                "the dataset do not match."
            )
        if dataset.base_pred != ckpt["base_pred"]:
            raise RemoraError(
                "The base_pred flag does not match between the pre-trained "
                "model and the dataset."
            )

    if ext_val:
        if ext_val_names is None:
            ext_val_names = [f"e_val_{idx}" for idx in range(len(ext_val))]
        else:
            assert len(ext_val_names) == len(ext_val)
        ext_sets = []
        for e_name, e_path in zip(ext_val_names, ext_val):
            ext_val_set = RemoraDataset.load_from_file(
                e_path.strip(),
                batch_size=batch_size,
                shuffle_on_iter=False,
                drop_last=False,
                drop_read_attrs=True,
            )
            if ext_val_set.mod_long_names != dataset.mod_long_names:
                ext_val_set.add_fake_base(
                    dataset.mod_long_names, dataset.mod_bases
                )
            ext_val_set.trim_kmer_context_bases(kmer_context_bases)
            ext_val_set.trim_chunk_context(chunk_context)
            ext_sets.append((e_name, ext_val_set))

    kmer_dim = int((sum(dataset.kmer_context_bases) + 1) * 4)
    test_input_sig = torch.randn(batch_size, 1, sum(dataset.chunk_context))
    test_input_seq = torch.randn(
        batch_size, kmer_dim, sum(dataset.chunk_context)
    )
    macs, params = profile(
        model, inputs=(test_input_sig, test_input_seq), verbose=False
    )
    LOGGER.info(
        f" Params (k) {params / (1000):.2f} | MACs (M) {macs / (1000 ** 2):.2f}"
    )

    LOGGER.info("Preparing training settings")
    criterion = torch.nn.CrossEntropyLoss()
    model = model.to(device)
    criterion = criterion.to(device)
    opt = load_optimizer(optimizer, model, lr, weight_decay)

    scheduler = select_scheduler(scheduler_name, opt, lr_sched_kwargs)

    label_counts = dataset.get_label_counts()
    LOGGER.info(f"Label distribution: {label_counts}")
    if len(label_counts) <= 1:
        raise RemoraError(
            "One or fewer output labels found. Ensure --focus-offset and "
            "--mod are specified correctly"
        )
    LOGGER.debug("Splitting dataset")
    trn_ds, val_ds = dataset.split_data(val_prop=val_prop, stratified=True)
    LOGGER.debug("Extracting head of train dataset")
    val_trn_ds = trn_ds.head(
        prop=val_prop, shuffle_on_iter=False, drop_last=False, stratified=True
    )
    if batch_label_weights is not None:
        LOGGER.debug("Applying batch label proportions")
        trn_ds.batch_label_props = np.array(batch_label_weights) / sum(
            batch_label_weights
        )
        trn_ds.order_chunks_by_batch_labels()
    LOGGER.info(f"Train label distribution: {trn_ds.get_label_counts()}")
    LOGGER.info(
        f"Held-out validation label distribution: {val_ds.get_label_counts()}"
    )
    LOGGER.info(
        "Training set validation label distribution: "
        f"{val_trn_ds.get_label_counts()}"
    )

    LOGGER.info("Running initial validation")
    # assess accuracy before first iteration
    val_metrics = val_fp.validate_model(
        model, dataset.mod_bases, criterion, val_ds, filt_frac
    )
    trn_metrics = val_fp.validate_model(
        model,
        dataset.mod_bases,
        criterion,
        val_trn_ds,
        filt_frac,
        "trn",
    )

    if ext_val:
        best_alt_val_accs = dict((e_name, 0) for e_name, _ in ext_sets)
        for e_name, e_set in ext_sets:
            val_fp.validate_model(
                model,
                dataset.mod_bases,
                criterion,
                e_set,
                filt_frac,
                e_name,
            )

    LOGGER.info("Start training")
    ebar = tqdm(
        total=epochs,
        smoothing=0,
        desc="Epochs",
        dynamic_ncols=True,
        position=0,
        leave=True,
        disable=os.environ.get("LOG_SAFE", False),
    )
    pbar = tqdm(
        total=len(trn_ds),
        desc="Epoch Progress",
        dynamic_ncols=True,
        position=1,
        leave=True,
        bar_format="{desc}: {percentage:3.0f}%|{bar}| " "{n_fmt}/{total_fmt}",
        disable=os.environ.get("LOG_SAFE", False),
    )
    ebar.set_postfix(
        acc_val=f"{val_metrics.acc:.4f}",
        acc_train=f"{trn_metrics.acc:.4f}",
        loss_val=f"{val_metrics.loss:.6f}",
        loss_train=f"{trn_metrics.loss:.6f}",
    )
    atexit.register(pbar.close)
    atexit.register(ebar.close)

    ckpt_save_data = {
        "epoch": 0,
        "state_dict": model.state_dict(),
        "opt": opt.state_dict(),
        "model_path": copy_model_path,
        "model_params": model_params,
        "chunk_context": dataset.chunk_context,
        "fixed_seq_len_chunks": model._variable_width_possible,
        "motifs": dataset.motifs,
        "num_motifs": dataset.num_motifs,
        "reverse_signal": dataset.reverse_signal,
        "mod_bases": dataset.mod_bases,
        "mod_long_names": dataset.mod_long_names,
        "base_pred": dataset.base_pred,
        "kmer_context_bases": dataset.kmer_context_bases,
        "base_start_justify": dataset.base_start_justify,
        "offset": dataset.offset,
        "model_version": constants.MODEL_VERSION,
        **dataset.sig_map_refiner.get_save_kwargs(),
    }
    bb, ab = dataset.kmer_context_bases
    best_val_acc = 0
    early_stop_epochs = 0
    breached = False
    for epoch in range(epochs):
        model.train()
        pbar.n = 0
        pbar.refresh()
        for epoch_i, ((sigs, seqs, seq_maps, seq_lens), labels, _) in enumerate(
            trn_ds
        ):
            sigs = torch.from_numpy(sigs)
            enc_kmers = torch.from_numpy(
                encoded_kmers.compute_encoded_kmer_batch(
                    bb, ab, seqs, seq_maps, seq_lens
                )
            )
            labels = torch.from_numpy(labels)
            sigs = sigs.to(device)
            enc_kmers = enc_kmers.to(device)
            outputs = model(sigs, enc_kmers)

            if high_conf_incorrect_thr_frac is None:
                labels = labels.to(device)
                loss = criterion(outputs, labels)
            else:
                batch_size = outputs.shape[0]
                conf_thresh, max_frac_skip = high_conf_incorrect_thr_frac
                max_nr_skip = int(np.floor(batch_size * max_frac_skip))
                preds = (
                    torch.nn.functional.softmax(outputs, dim=1).detach().cpu()
                )
                highest_preds, high_conf_cl = torch.max(preds, dim=1)
                cl_match = labels == high_conf_cl
                # if there are more mismatches than the maximum number allowed
                # to be skipped, set confidence threshold to allow no more than
                # max_nr_skip items be skipped
                if batch_size - cl_match.sum() > max_nr_skip:
                    mm_preds = highest_preds[~cl_match]
                    mm_preds.sort(descending=True)
                    conf_thresh = max(conf_thresh, mm_preds[max_nr_skip])
                mask = cl_match.logical_or(highest_preds < conf_thresh)
                # avoid sending labels to device until after above computations
                labels = labels.to(device)
                loss = criterion(outputs[mask], labels[mask])

            opt.zero_grad()
            loss.backward()
            opt.step()
            batch_fp.write(
                f"{(epoch * len(trn_ds)) + epoch_i}\t"
                f"{loss.detach().cpu():.6f}"
            )
            if high_conf_incorrect_thr_frac is None:
                batch_fp.write("\n")
            else:
                batch_fp.write(f"\t{batch_size - mask.sum()}\n")

            pbar.update()
            pbar.refresh()

        niter = (epoch + 1) * len(trn_ds)
        val_metrics = val_fp.validate_model(
            model,
            dataset.mod_bases,
            criterion,
            val_ds,
            filt_frac,
            nepoch=epoch + 1,
            niter=niter,
            disable_pbar=True,
        )
        trn_metrics = val_fp.validate_model(
            model,
            dataset.mod_bases,
            criterion,
            val_trn_ds,
            filt_frac,
            "trn",
            nepoch=epoch + 1,
            niter=niter,
            disable_pbar=True,
        )

        scheduler.step()

        if breached:
            if val_metrics.acc <= REGRESSION_THRESHOLD:
                LOGGER.warning("Remora training unstable")
        else:
            if val_metrics.acc >= BREACH_THRESHOLD:
                breached = True
                LOGGER.debug(
                    f"{BREACH_THRESHOLD * 100}% accuracy threshold surpassed"
                )

        if val_metrics.acc > best_val_acc:
            best_val_acc = val_metrics.acc
            early_stop_epochs = 0
            LOGGER.debug(
                f"Saving best model after {epoch + 1} epochs with "
                f"val_acc {val_metrics.acc}"
            )
            save_model(
                model,
                ckpt_save_data,
                out_path,
                epoch,
                opt,
            )
        else:
            early_stop_epochs += 1

        if ext_val:
            for e_name, e_set in ext_sets:
                e_val_metrics = val_fp.validate_model(
                    model,
                    dataset.mod_bases,
                    criterion,
                    e_set,
                    filt_frac,
                    e_name,
                    nepoch=epoch + 1,
                    niter=niter,
                    disable_pbar=True,
                )
                if e_val_metrics.acc <= best_alt_val_accs[e_name]:
                    continue
                best_alt_val_accs[e_name] = e_val_metrics.acc
                early_stop_epochs = 0
                LOGGER.debug(
                    f"Saving best model based on {e_name} "
                    f"validation set after {epoch + 1} epochs "
                    f"with val_acc {e_val_metrics.acc}"
                )
                save_model(
                    model,
                    ckpt_save_data,
                    out_path,
                    epoch,
                    opt,
                    model_name=f"model_ext_val_{e_name}_best.checkpoint",
                    model_name_torchscript=f"model_ext_val_{e_name}_best.pt",
                )

        if int(epoch + 1) % save_freq == 0:
            save_model(
                model,
                ckpt_save_data,
                out_path,
                epoch,
                opt,
                model_name=f"model_{epoch + 1:06d}.checkpoint",
                model_name_torchscript=f"model_{epoch + 1:06d}.pt",
            )

        ebar.set_postfix(
            acc_val=f"{val_metrics.acc:.4f}",
            acc_train=f"{trn_metrics.acc:.4f}",
            loss_val=f"{val_metrics.loss:.6f}",
            loss_train=f"{trn_metrics.loss:.6f}",
        )
        ebar.update()
        if early_stopping and early_stop_epochs >= early_stopping:
            LOGGER.info(
                "No validation accuracy improvement after"
                f" {early_stopping} epoch(s). Stopping training early."
            )
            break
    ebar.close()
    pbar.close()
    LOGGER.info("Saving final model checkpoint")
    save_model(
        model,
        ckpt_save_data,
        out_path,
        epoch,
        opt,
        model_name=constants.FINAL_MODEL_FILENAME,
        model_name_torchscript=constants.FINAL_TORCHSCRIPT_MODEL_FILENAME,
    )
    LOGGER.info("Training complete")


if __name__ == "__main__":
    NotImplementedError("This is a module.")



import torch


def swish(x):
    """Swish activation

    Swish is self-gated linear activation :math:`x sigma(x)`

    For details see: https://arxiv.org/abs/1710.05941

    Note:
        Original definition has a scaling parameter for the gating value,
        making it a generalisation of the (logistic approximation to) the GELU.
        Evidence presented, e.g. https://arxiv.org/abs/1908.08681 that swish-1
        performs comparable to tuning the parameter.

    """
    return x * torch.sigmoid(x)
